import numpy as np
import pandas as pd
import collections

from . import ops
from .core.specs import _get_default_colnames, _verify_columns

__all__ = [
    "make_chromarms",
    "binnify",
    "digest",
    "frac_mapped",
    "frac_gc",
    "frac_gene_coverage",
    "pair_by_distance",
]


def make_chromarms(
    chromsizes,
    midpoints,
    cols_chroms=("chrom", "length"),
    cols_mids=("chrom", "mid"),
    suffixes=("_p", "_q"),
):
    """
    Split chromosomes into chromosome arms.

    Parameters
    ----------
    chromsizes : pandas.Dataframe or pandas.Series
        If pandas.Series, a map from chromosomes to lengths in bp.
        If pandas.Dataframe, a dataframe with columns defined by cols_chroms.
        If cols_chroms is a triplet (e.g. 'chrom','start','end'), then
        values in chromsizes[cols_chroms[1]].values must all be zero.

    midpoints : pandas.Dataframe or dict-like
        Mapping of chromosomes to midpoint (aka centromere) locations.
        If pandas.Series, a map from chromosomes to midpoints in bp.
        If pandas.Dataframe, a dataframe with columns defined by cols_mids.

    cols_chroms : (str, str) or (str, str, str)
        Two columns

    suffixes : tuple, optional
        Suffixes to name chromosome arms. Defaults to p and q.

    Returns
    -------
    df_chromarms
        4-column BED-like DataFrame (chrom, start, end, name).
        Arm names are chromosome names + suffix.
        Any chromosome not included in ``mids`` will be not be split.

    """
    columns_to_drop = ["index", "sub_index_"]
    if len(cols_chroms) == 2:
        ck1, sk1 = cols_chroms
    elif len(cols_chroms) == 3:
        ck1, sk1, ek1 = cols_chroms

    if isinstance(chromsizes, pd.Series):
        df_chroms = (
            pd.DataFrame(chromsizes).reset_index().rename(columns={"index": ck1})
        )
    elif isinstance(chromsizes, pd.DataFrame):
        df_chroms = chromsizes.copy()
    else:
        raise ValueError("unknown input type for chromsizes")

    if len(cols_chroms) == 2:
        _verify_columns(df_chroms, [ck1, sk1])
        columns_to_drop += [sk1]
        df_chroms["end"] = df_chroms[sk1].values
        df_chroms["start"] = 0
        sk1, ek1 = "start", "end"
    elif len(cols_chroms) == 3:
        ck1, sk1, ek1 = cols_chroms
        _verify_columns(df_chroms, [ck1, sk1, ek1])
        if any((df_chroms[sk1].values != 0)):
            raise ValueError("all values in starts column must be zero")
    else:
        raise ValueError("invalid number of cols_chroms")

    ck2, sk2 = cols_mids
    if isinstance(midpoints, dict):
        df_mids = pd.DataFrame.from_dict(midpoints, orient="index", columns=[sk2])
        df_mids.reset_index(inplace=True)
        df_mids.rename(columns={"index": ck2}, inplace=True)
    elif isinstance(midpoints, pd.DataFrame):
        df_mids = midpoints.copy()
    else:
        raise ValueError("unknown input type for midpoints")
    _verify_columns(df_mids, [ck2, sk2])
    df_mids["start"] = df_mids[sk2]
    df_mids["end"] = df_mids[sk2]

    df_chromarms = ops.subtract(
        df_chroms,
        df_mids,
        cols1=(ck1, sk1, ek1),
        cols2=(ck2, "start", "end"),
        return_index=True,
    )
    if df_chromarms["sub_index_"].max() > 1:
        raise ValueError(
            "chromosome split into more than two arms, double-check midpoints"
        )
    df_chromarms["name"] = df_chromarms[ck1] + [
        suffixes[i] for i in df_chromarms["sub_index_"].values
    ]
    # df_chromarms.drop(columns=columns_to_drop, inplace=True)
    return df_chromarms[[ck1, sk1, ek1, "name"]]


def binnify(chromsizes, binsize, rel_ids=False):
    """
    Divide a genome into evenly sized bins.

    Parameters
    ----------
    chromsizes : Series
        pandas Series indexed by chromosome name with chromosome lengths in bp.

    binsize : int
        size of bins in bp

    Returns
    -------
    bintable : pandas.DataFrame with columns: 'chrom', 'start', 'end'.

    """

    if not isinstance(binsize, int):
        raise ValueError("binsize must be int")

    def _each(chrom):
        clen = chromsizes[chrom]
        n_bins = int(np.ceil(clen / binsize))
        binedges = np.arange(0, (n_bins + 1)) * binsize
        binedges[-1] = clen
        return pd.DataFrame(
            {"chrom": [chrom] * n_bins, "start": binedges[:-1], "end": binedges[1:]},
            columns=["chrom", "start", "end"],
        )

    bintable = pd.concat(map(_each, chromsizes.keys()), axis=0, ignore_index=True)

    if rel_ids:
        bintable["rel_id"] = bintable.groupby("chrom").cumcount()

    # if as_cat:
    #     bintable['chrom'] = pd.Categorical(
    #         bintable['chrom'],
    #         categories=list(chromsizes.keys()),
    #         ordered=True)

    return bintable


def digest(fasta_records, enzyme):
    """
    Divide a genome into restriction fragments.

    Parameters
    ----------
    fasta_records : OrderedDict
        Dictionary of chromosome names to sequence records.
        Created by: bioframe.load_fasta('/path/to/fasta.fa')

    enzyme: str
        Name of restriction enzyme.

    Returns
    -------
    Dataframe with columns: 'chrom', 'start', 'end'.

    """
    try:
        import Bio.Restriction as biorst
        import Bio.Seq as bioseq
    except ImportError:
        raise ImportError("Biopython is required to use digest")

    # http://biopython.org/DIST/docs/cookbook/Restriction.html#mozTocId447698
    if not isinstance(fasta_records, dict):
        raise ValueError(
            "fasta records must be provided as an OrderedDict, can be created by bioframe.load_fasta"
        )
    chroms = fasta_records.keys()
    try:
        cut_finder = getattr(biorst, enzyme).search
    except AttributeError:
        raise ValueError("Unknown enzyme name: {}".format(enzyme))

    def _each(chrom):
        seq = bioseq.Seq(str(fasta_records[chrom][:]))
        cuts = np.r_[0, np.array(cut_finder(seq)) + 1, len(seq)].astype(int)
        n_frags = len(cuts) - 1

        frags = pd.DataFrame(
            {"chrom": [chrom] * n_frags, "start": cuts[:-1], "end": cuts[1:]},
            columns=["chrom", "start", "end"],
        )
        return frags

    return pd.concat(map(_each, chroms), axis=0, ignore_index=True)


def frac_mapped(df, fasta_records, return_input=True):
    """
    Calculate the fraction of mapped base-pairs for each interval in a dataframe.

    Parameters
    ----------
    df : pandas.DataFrame
        A sets of genomic intervals stored as a DataFrame.

    fasta_records : OrderedDict
        Dictionary of chromosome names to sequence records.
        Created by: bioframe.load_fasta('/path/to/fasta.fa')

    return_input: bool
        if False, only return Series named frac_mapped.

    Returns
    -------
    df_mapped : pd.DataFrame
        Original dataframe with new column 'frac_mapped' appended.

    """

    if not set(df["chrom"].values).issubset(set(fasta_records.keys())):
        raise ValueError(
            "chrom from intervals not in fasta_records: double-check genome agreement"
        )
    if not isinstance(fasta_records, dict):
        raise ValueError(
            "fasta records must be provided as an OrderedDict, can be created by bioframe.load_fasta"
        )

    def _each(bin):
        s = str(fasta_records[bin.chrom][bin.start : bin.end])
        nbases = len(s)
        n = s.count("N")
        n += s.count("n")
        return (nbases - n) / nbases if nbases > 0 else 0

    if return_input:
        return pd.concat(
            [df, df.apply(_each, axis=1).rename("frac_mapped", inplace=True)],
            axis="columns",
        )
    else:
        return df.apply(_each, axis=1).rename("frac_mapped", inplace=True)


def frac_gc(df, fasta_records, mapped_only=True, return_input=True):
    """
    Calculate the fraction of GC basepairs for each interval in a dataframe.

    Parameters
    ----------
    df : pandas.DataFrame
        A sets of genomic intervals stored as a DataFrame.

    fasta_records : OrderedDict
        Dictionary of chromosome names to sequence records.
        Created by: bioframe.load_fasta('/path/to/fasta.fa')

    mapped_only: bool
        if True, ignore 'N' in the fasta_records for calculation.
        if True and there are no mapped base-pairs in an interval, return np.nan.

    return_input: bool
        if False, only return Series named frac_mapped.

    Returns
    -------
    df_mapped : pd.DataFrame
        Original dataframe with new column 'frac_mapped' appended.

    """
    if not set(df["chrom"].values).issubset(set(fasta_records.keys())):
        raise ValueError(
            "chrom from intervals not in fasta_records: double-check genome agreement"
        )
    if not isinstance(fasta_records, dict):
        raise ValueError(
            "fasta records must be provided as an OrderedDict, can be created by bioframe.load_fasta"
        )

    def _each(chrom_group):
        chrom = chrom_group.name
        seq = fasta_records[chrom]
        seq = str(seq[:])
        gc = []
        for _, bin in chrom_group.iterrows():
            s = seq[bin.start : bin.end]
            g = s.count("G")
            g += s.count("g")
            c = s.count("C")
            c += s.count("c")
            nbases = len(s)
            if mapped_only:
                n = s.count("N")
                n += s.count("n")
                nbases -= n
            gc.append((g + c) / nbases if nbases > 0 else np.nan)
        return gc

    out = df.groupby("chrom", sort=False).apply(_each)

    if return_input:
        return pd.concat(
            [df, pd.Series(data=np.concatenate(out), index=df.index).rename("GC")],
            axis="columns",
        )
    else:
        return pd.Series(data=np.concatenate(out), index=df.index).rename("GC")


def frac_gene_coverage(df, ucsc_mrna):
    """
    Calculate number and fraction of overlaps by predicted and verified 
    RNA isoforms for a set of intervals stored in a dataframe.

    Parameters
    ----------
    df : pd.DataFrame
        Set of genomic intervals stored as a dataframe.

    ucsc_mrna: str or DataFrame
        Name of UCSC genome or all_mrna.txt dataframe from UCSC or similar.

    Returns
    -------
    df_gene_coverage : pd.DataFrame

    """
    if isinstance(ucsc_mrna, str):
        from .io.resources import UCSCClient

        mrna = UCSCClient(ucsc_mrna).fetch_mrna()
    else:
        mrna = ucsc_mrna

    mrna = mrna.rename(columns={"tName": "chrom", "tStart": "start", "tEnd": "end"})
    df_gene_coverage = ops.coverage(df, mrna)
    df_gene_coverage = ops.count_overlaps(df_gene_coverage, mrna)

    return df_gene_coverage


def pair_by_distance(
    df,
    min_sep,
    max_sep,
    min_intervening=None,
    max_intervening=None,
    relative_to="midpoints",
    cols=None,
    suffixes=("_1", "_2"),
):
    """
    From a dataframe of genomic intervals, find all unique pairs of intervals
    that are between ``min_sep`` and ``max_sep`` bp separated from each other.

    Parameters
    ----------
    df : pandas.DataFrame
        A BED-like dataframe.
    min_sep, max_sep : int
        Minimum and maximum separation between intervals in bp.
        Min > 0 and Max >= Min.
    min_intervening, max_intervening : int
        Minimum and maximum number of intervening intervals separating pairs.
        Min > 0 and Max >= Min.
    relative_to : str,
        Whether to calculate distances between interval "midpoints" or "endpoints".
        Default "midpoints".
    cols : (str, str, str) or None
        The names of columns containing the chromosome, start and end of the
        genomic intervals, provided separately for each set. The default
        values are 'chrom', 'start', 'end'.
    suffixes : (str, str), optional
        The column name suffixes for the two interval sets in the output.
        The first interval of each output pair is always upstream of the
        second.

    Returns
    -------
    pandas.DataFrame
        A BEDPE-like dataframe of paired intervals from ``df``.

    """
    ck, sk, ek = _get_default_colnames() if cols is None else cols

    # Sort intervals by genomic coordinates
    df = df.sort_values([ck, sk, ek]).reset_index(drop=True)

    if min_sep >= max_sep:
        raise ValueError("min_sep must be less than max_sep")
    if min_sep < 0:
        raise ValueError("min_sep must be >=0")

    if min_intervening is None:
        min_intervening = 0
    if max_intervening is None:
        max_intervening = df.index.max()
    if min_intervening > max_intervening:
        raise ValueError("min_intervening must be less or equal to max_intervening")
    if min_intervening < 0:
        raise ValueError("min_intervening must be >=0")

    mids = (df[sk] + df[ek]) // 2

    # For each interval, generate a probe interval on its right
    if relative_to == "endpoints":
        print("endpoint")
        ref = df[ek]
    elif relative_to == "midpoints":
        ref = mids
    else:
        raise ValueError("relative_to must either specify 'midpoints' or 'endpoints' ")
    right_probe = df[[ck]].copy()
    right_probe[sk] = ref + min_sep // 2
    right_probe[ek] = ref + (max_sep + 1) // 2

    # For each interval, also generate a probe interval on its left
    if relative_to == "endpoints":
        ref = df[sk]
    elif relative_to == "midpoints":
        ref = mids
    else:
        raise ValueError("relative_to must either specify 'midpoints' or 'endpoints' ")
    left_probe = df[[ck]].copy()
    left_probe[sk] = ref - max_sep // 2
    left_probe[ek] = ref - (min_sep + 1) // 2

    # Intersect right-handed probes (from intervals on the left)
    # with left-handed probes (from intervals on the right)
    idxs = ops.overlap(
        right_probe,
        left_probe,
        suffixes=suffixes,
        how="inner",
        return_index=True,
        return_input=False,
    )

    # Select only the pairs that are separated by
    # at least min_intervening intervals and no more than max_intervening intervals
    idxs["intervening"] = (
        np.abs(idxs[f"index{suffixes[0]}"] - idxs[f"index{suffixes[1]}"]) - 1
    )
    idxs = idxs.query(
        f"intervening<={max_intervening} and intervening>={min_intervening}"
    )

    left_ivals = (
        df.iloc[idxs[f"index{suffixes[0]}"].values]
        .rename(columns=lambda x: x + suffixes[0])
        .reset_index(drop=True)
    )
    right_ivals = (
        df.iloc[idxs[f"index{suffixes[1]}"].values]
        .rename(columns=lambda x: x + suffixes[1])
        .reset_index(drop=True)
    )

    return pd.concat([left_ivals, right_ivals], axis=1)



import pandas as pd
import numpy as np
from .specs import _get_default_colnames, _verify_columns, is_chrom_dtype
from .stringops import parse_region_string, to_ucsc_string, is_complete_ucsc_string
from . import checks

__all__ = [
    "from_dict",
    "from_series",
    "from_list",
    "from_any",
    "make_viewframe",
    "sanitize_bedframe",
]

### conversions from various input formats into dataframes ###


def from_dict(regions, cols=None):
    """
    Makes a dataframe from a dictionary of {str,int} pairs, interpreted as chromosome names.

    Note that {str,(int,int)} dictionaries of tuples are no longer supported!

    Parameters
    ----------

    regions : dict

    name_col : str
        Default 'name'.

    cols : (str, str, str) or None
        The names of columns containing the chromosome, start and end of the
        genomic intervals, provided separately for each set. The default
        values are 'chrom', 'start', 'end'.

    Returns
    -------
    df : pandas.DataFrame
    """
    ck1, sk1, ek1 = _get_default_colnames() if cols is None else cols
    data = []
    for k, v in dict(regions).items():
        chrom = k
        if np.isscalar(v):
            start = 0
            end = v
        else:
            raise ValueError("Unsupported dict format: {type(v)}")
        data.append([chrom, start, end])
    return pd.DataFrame(data, columns=[ck1, sk1, ek1])


def from_series(regions, cols=None):
    ck1, sk1, ek1 = _get_default_colnames() if cols is None else cols
    chroms = regions.index.values
    data = {ck1: chroms, sk1: 0, ek1: regions.values}
    return pd.DataFrame(data)


def from_list(regions, name_col="name", cols=None):
    ck1, sk1, ek1 = _get_default_colnames() if cols is None else cols
    df = pd.DataFrame(regions)
    if df.shape[1] == 3:
        df.columns = [ck1, sk1, ek1]
    elif df.shape[1] == 4:
        df.columns = [ck1, sk1, ek1, name_col]
    else:
        raise ValueError("wrong number of columns for list input format")
    return df


def from_ucsc_string_list(region_list, cols=None):
    ck1, sk1, ek1 = _get_default_colnames() if cols is None else cols
    parsed = [parse_region_string(i) for i in region_list]
    df = pd.DataFrame(parsed, columns=[ck1, sk1, ek1])
    return df


def from_any(regions, fill_null=False, name_col="name", cols=None):
    """
    Attempts to make a genomic interval dataframe with columns [chr, start, end, name_col] from a variety of input types.

    Parameters
    ----------
    regions : supported input
        Currently supported inputs:

            - dataframe
            - series of UCSC strings
            - dictionary of {str:int} key value pairs
            - pandas series where the index is interpreted as chromosomes and values are interpreted as end
            - list of tuples or lists, either [(chrom,start,end)] or [(chrom,start,end,name)]
            - tuple of tuples or lists, either [(chrom,start,end)] or [(chrom,start,end,name)]

    fill_null : False or dictionary
        Accepts a dictionary of {str:int} pairs, interpreted as chromosome sizes.
        Kept or backwards compatibility. Default False.

    name_col : str
        Column name. Only used if 4 column list is provided. Default "name".

    cols : (str,str,str)
        Names for dataframe columns.
        Default None sets them with get_default_colnames().

    Returns
    -------
    out_df:dataframe

    """
    ck1, sk1, ek1 = _get_default_colnames() if cols is None else cols

    if type(regions) is pd.core.frame.DataFrame:
        if set([ck1, sk1, ek1]).issubset(regions.columns):
            out_df = regions.copy()
        elif (len(regions[name_col].values.shape) == 1) and is_complete_ucsc_string(
            regions[name_col].values[0]
        ):
            out_df = from_ucsc_string_list(
                regions[name_col].values, cols=[ck1, sk1, ek1]
            )
        else:
            raise ValueError("Unknown dataFrame format: check column names")

    elif type(regions) is dict:
        out_df = from_dict(regions, cols=[ck1, sk1, ek1])

    elif type(regions) is pd.core.series.Series:
        out_df = from_series(regions, cols=[ck1, sk1, ek1])

    elif type(regions) is tuple:
        if np.shape(regions) == (3,):
            out_df = from_list([regions], name_col=name_col, cols=[ck1, sk1, ek1])

        elif len(np.shape(regions)) == 1 and type(regions[0]) is str:
            out_df = from_ucsc_string_list(regions, cols=[ck1, sk1, ek1])
        else:
            out_df = from_list(list(regions), name_col=name_col, cols=[ck1, sk1, ek1])

    elif type(regions) is list:
        if np.shape(regions) == (3,):
            out_df = from_list([regions], name_col=name_col, cols=[ck1, sk1, ek1])
        elif len(np.shape(regions)) == 1 and type(regions[0]) is str:
            out_df = from_ucsc_string_list(regions, cols=[ck1, sk1, ek1])
        else:
            out_df = from_list(regions, name_col=name_col, cols=[ck1, sk1, ek1])
    else:
        raise ValueError(f"Unknown input format: {type(regions)}")

    if fill_null:
        try:
            out_df[sk1].fillna(0, inplace=True)
            ends = []
            for i in range(len(out_df)):
                if out_df[ek1].values[i] is None:
                    ends.append(fill_null[out_df[ck1].values[i]])
                else:
                    ends.append(out_df[ek1].values[i])
            out_df[ek1] = ends
        except:
            raise ValueError("could not fill ends with provided chromsizes")

    return out_df


def add_ucsc_name_column(reg_df, name_col="name", cols=None):
    """
    Auto-creates a UCSC name 'chrom:start-end' for each region (chrom,start,end) in reg_df.

    Replaces name_col if it exists.



    """
    ck1, sk1, ek1 = _get_default_colnames() if cols is None else cols
    df = reg_df.copy()
    _verify_columns(df, [ck1, sk1, ek1])
    data = zip(df[ck1], df[sk1], df[ek1])
    df[name_col] = [to_ucsc_string(i) for i in data]
    return df


def make_viewframe(
    regions,
    check_bounds=None,
    name_style=None,
    view_name_col="name",
    cols=None,
):
    """
    Makes and validates a dataframe `view_df` out of regions.

    Parameters
    ----------
    regions : supported input type
        Currently supported input types:

            - a dictionary where keys are strings and values are integers {str:int}, specifying regions (chrom, 0, end, chrom)
            - a pandas series of chromosomes lengths with index specifying region names
            - a list of tuples [(chrom,start,end), ...] or [(chrom,start,end,name), ...]
            - a pandas DataFrame, skips to validation step

    name_style : None or "ucsc"
        If None and no column view_name_col, propagate values from cols[0]
        If "ucsc" and no column view_name_col, create UCSC style names

    check_bounds : None, or chromosome sizes provided as any of valid formats above
        Optional, if provided checks if regions in the view are contained by regions
        supplied in check_bounds, typically provided as a series of chromosome sizes.
        Default None.

    view_name_col : str
        Specifies column name of the view regions. Default 'name'.

    cols : (str, str, str) or None
        The names of columns containing the chromosome, start and end of the
        genomic intervals, provided separately for each set. The default
        values are 'chrom', 'start', 'end'.

    Returns
    -------
    view_df:dataframe satisfying properties of a view

    """
    ck1, sk1, ek1 = _get_default_colnames() if cols is None else cols

    view_df = from_any(regions, name_col=view_name_col, cols=cols)

    if check_bounds is not None:
        bounds_df = from_any(check_bounds, name_col="bounds", cols=cols)
        if not checks.is_contained(
            view_df,
            bounds_df,
            df_view_col=None,
            view_name_col="bounds",
            cols=cols,
        ):
            raise ValueError(
                "Invalid input to make a viewFrame, regions not contained by bounds"
            )

    if not view_name_col in view_df.columns:
        if name_style is None:
            view_df[view_name_col] = view_df[ck1].values
        elif name_style.lower() == "ucsc":
            view_df = add_ucsc_name_column(view_df, name_col=view_name_col, cols=cols)
        else:
            raise ValueError("unknown value for name_style")

    if checks.is_viewframe(
        view_df, view_name_col=view_name_col, cols=cols, raise_errors=True
    ):
        return view_df
    else:
        raise ValueError("could not make valid viewFrame, retry with new input")


def sanitize_bedframe(
    df1,
    recast_dtypes=True,
    drop_null=False,
    start_exceed_end_action=None,
    cols=None,
):
    """
    Attempts to clean a genomic interval dataframe to be a valid bedframe.

    Parameters
    ----------
    df1 : pandas.DataFrame

    recast_dtypes : bool
        Whether to attempt to recast column dtypes to pandas nullable dtypes.

    drop_null : bool
        Drops rows with pd.NA. Default False.

    start_exceed_end_action : str or None
        Options: 'flip' or 'drop' or None. Default None.

            - If 'flip', attempts to sanitize by flipping intervals with start>end.
            - If 'drop' attempts to sanitize dropping intervals with start>end.
            - If None, does not alter these intervals if present.

    cols : (str, str, str) or None
        The names of columns containing the chromosome, start and end of the
        genomic intervals, provided separately for each set. The default
        values are 'chrom', 'start', 'end'.

    Returns
    -------
    out_df : pandas.DataFrame
        Sanitized dataframe satisfying the properties of a bedframe.

    Notes
    ------
    The option ``start_exceed_end_action='flip'`` may be useful for gff files with strand information but starts > ends.

    """
    ck1, sk1, ek1 = _get_default_colnames() if cols is None else cols

    out_df = df1.copy()

    _verify_columns(out_df, [ck1, sk1, ek1])

    if recast_dtypes:
        chrom_dtype, start_dtype, end_dtype = out_df.dtypes[[ck1, sk1, ek1]]
        if not is_chrom_dtype(chrom_dtype):
            out_df[ck1] = out_df[ck1].astype(str)
        if not ((start_dtype is pd.Int64Dtype()) and (end_dtype is pd.Int64Dtype())):
            out_df[sk1] = out_df[sk1].astype(pd.Int64Dtype())
            out_df[ek1] = out_df[ek1].astype(pd.Int64Dtype())

    nan_intervals = pd.isnull(out_df[[ck1, sk1, ek1]]).any(axis=1)
    out_df.loc[nan_intervals, [ck1, sk1, ek1]] = pd.NA
    if drop_null:
        out_df.dropna(axis=0, inplace=True)
        out_df.reset_index(drop=True, inplace=True)

    if start_exceed_end_action is not None:
        start_exceed_end_action = start_exceed_end_action.lower()
        if ((out_df[ek1] - out_df[sk1]) < 0).any():
            inds = ((out_df[ek1] - out_df[sk1]) < 0).values
            if start_exceed_end_action == "drop":
                out_df = out_df.loc[inds == 0]
            elif start_exceed_end_action == "flip":
                out_df.loc[inds, [sk1, ek1]] = out_df.loc[inds, [ek1, sk1]].values
            else:
                raise ValueError("unknown action for intervals with start>end")
            out_df.reset_index(drop=True, inplace=True)

    if checks.is_bedframe(out_df, cols=cols):
        return out_df
    else:
        raise ValueError("could not sanitize")



import collections
import pandas as pd
import numpy as np

__all__ = [
    "update_default_colnames",
    "is_chrom_dtype",
]

_rc = {"colnames": {"chrom": "chrom", "start": "start", "end": "end"}}


def _get_default_colnames():
    """
    Returns default column names.

    These defaults be updated with :func:`update_default_colnames`.

    Returns
    -------
    colnames : triplet (str, str, str)

    """
    return _rc["colnames"]["chrom"], _rc["colnames"]["start"], _rc["colnames"]["end"]


class update_default_colnames:
    def __init__(self, new_colnames):
        self._old_colnames = dict(_rc["colnames"])
        if isinstance(new_colnames, collections.abc.Iterable):
            if len(new_colnames) != 3:
                raise ValueError(
                    "Please, specify new columns using a list of "
                    "3 strings or a dict!"
                )
            (
                _rc["colnames"]["chrom"],
                _rc["colnames"]["start"],
                _rc["colnames"]["end"],
            ) = new_colnames
        elif isinstance(new_colnames, collections.abc.Mapping):
            _rc["colnames"].update(
                {
                    k: v
                    for k, v in new_colnames.items()
                    if k in ["chrom", "start", "end"]
                }
            )
        else:
            raise ValueError(
                "Please, specify new columns using a list of " "3 strings or a dict!"
            )

    def __enter__(self):
        return self

    def __exit__(self, *args):
        _rc["colnames"] = self._old_colnames


def _verify_columns(df, colnames, return_as_bool=False):
    """
    Raises ValueError if columns with colnames are not present in dataframe df.

    Parameters
    ----------
    df: pandas.DataFrame

    colnames: list of column names

    return_as_bool : bool
        If True, returns as a boolean instead of raising errors. Default False.

    """

    if not type(df) is pd.core.frame.DataFrame:
        if return_as_bool:
            return False
        raise ValueError("df is not a dataframe")

    if len(set(colnames)) < len(colnames):
        raise ValueError("column names must be unique")

    if not set(colnames).issubset(df.columns):
        if return_as_bool:
            return False
        raise ValueError(
            ", ".join(set(colnames).difference(set(df.columns)))
            + " not in keys of df.columns"
        )
    if return_as_bool:
        return True


def _verify_column_dtypes(df, cols=None, return_as_bool=False):
    """
    Checks that dataframe `df` has chrom, start, end columns with valid dtypes.
    Raises TypeErrors if cols have invalid dtypes.

    Parameters
    ----------
    df : pandas.DataFrame

    cols : (str, str, str) or None
        The names of columns containing the chromosome, start and end of the
        genomic intervals, provided separately for each set. The default
        values are 'chrom', 'start', 'end'.

    return_as_bool : bool
        If true, returns as a boolean instead of raising errors. Default False.

    """
    ck1, sk1, ek1 = _get_default_colnames() if cols is None else cols
    if not _verify_columns(df, [ck1, sk1, ek1], return_as_bool=True):
        if return_as_bool:
            return False
        raise ValueError("could not verify columns")

    chrom_dtype, start_dtype, end_dtype = df.dtypes[[ck1, sk1, ek1]]

    if not is_chrom_dtype(chrom_dtype):
        if return_as_bool:
            return False
        raise TypeError(
            "invalid df['chrom'] dtype, must be object, string, or categorical"
        )
    if not pd.api.types.is_integer_dtype(start_dtype):
        if return_as_bool:
            return False
        raise TypeError("invalid df['start'] dtype, must be integer")

    if not pd.api.types.is_integer_dtype(end_dtype):
        if return_as_bool:
            return False
        raise TypeError("invalid df['end'] dtype, must be integer")

    if return_as_bool:
        return True


def is_chrom_dtype(chrom_dtype):
    """
    Returns True if dtype is any of the allowed bioframe chrom dtypes, False otherwise.
    """
    return np.any(
        [
            pd.api.types.is_string_dtype(chrom_dtype),
            pd.api.types.is_object_dtype(chrom_dtype),
            pd.api.types.is_categorical_dtype(chrom_dtype),
        ]
    )



from io import StringIO

import pandas as pd
import numpy as np
import pytest

import bioframe

import bioframe.core.checks as checks
from bioframe.core.construction import make_viewframe

# import pyranges as pr

# def bioframe_to_pyranges(df):
#     pydf = df.copy()
#     pydf.rename(
#         {"chrom": "Chromosome", "start": "Start", "end": "End"},
#         axis="columns",
#         inplace=True,
#     )
#     return pr.PyRanges(pydf)


# def pyranges_to_bioframe(pydf):
#     df = pydf.df
#     df.rename(
#         {"Chromosome": "chrom", "Start": "start", "End": "end", "Count": "n_intervals"},
#         axis="columns",
#         inplace=True,
#     )
#     return df


# def pyranges_overlap_to_bioframe(pydf):
#     ## convert the df output by pyranges join into a bioframe-compatible format
#     df = pydf.df.copy()
#     df.rename(
#         {
#             "Chromosome": "chrom_1",
#             "Start": "start_1",
#             "End": "end_1",
#             "Start_b": "start_2",
#             "End_b": "end_2",
#         },
#         axis="columns",
#         inplace=True,
#     )
#     df["chrom_1"] = df["chrom_1"].values.astype("object")  # to remove categories
#     df["chrom_2"] = df["chrom_1"].values
#     return df


chroms = ["chr12", "chrX"]


def mock_bioframe(num_entries=100):
    pos = np.random.randint(1, 1e7, size=(num_entries, 2))
    df = pd.DataFrame()
    df["chrom"] = np.random.choice(chroms, num_entries)
    df["start"] = np.min(pos, axis=1)
    df["end"] = np.max(pos, axis=1)
    df.sort_values(["chrom", "start"], inplace=True)
    return df


############# tests #####################
def test_select():
    df1 = pd.DataFrame(
        [["chrX", 3, 8], ["chr1", 4, 5], ["chrX", 1, 5]],
        columns=["chrom", "start", "end"],
    )

    region1 = "chr1:4-10"
    df_result = pd.DataFrame([["chr1", 4, 5]], columns=["chrom", "start", "end"])
    pd.testing.assert_frame_equal(
        df_result, bioframe.select(df1, region1).reset_index(drop=True)
    )

    region1 = "chrX"
    df_result = pd.DataFrame(
        [["chrX", 3, 8], ["chrX", 1, 5]], columns=["chrom", "start", "end"]
    )
    pd.testing.assert_frame_equal(
        df_result, bioframe.select(df1, region1).reset_index(drop=True)
    )

    region1 = "chrX:4-6"
    df_result = pd.DataFrame(
        [["chrX", 3, 8], ["chrX", 1, 5]], columns=["chrom", "start", "end"]
    )
    pd.testing.assert_frame_equal(
        df_result, bioframe.select(df1, region1).reset_index(drop=True)
    )

    ### select with non-standard column names
    region1 = "chrX:4-6"
    new_names = ["chr", "chrstart", "chrend"]
    df1 = pd.DataFrame(
        [["chrX", 3, 8], ["chr1", 4, 5], ["chrX", 1, 5]],
        columns=new_names,
    )
    df_result = pd.DataFrame(
        [["chrX", 3, 8], ["chrX", 1, 5]],
        columns=new_names,
    )
    pd.testing.assert_frame_equal(
        df_result, bioframe.select(df1, region1, cols=new_names).reset_index(drop=True)
    )
    region1 = "chrX"
    pd.testing.assert_frame_equal(
        df_result, bioframe.select(df1, region1, cols=new_names).reset_index(drop=True)
    )

    ### select from a DataFrame with NaNs
    colnames = ["chrom", "start", "end", "view_region"]
    df = pd.DataFrame(
        [
            ["chr1", -6, 12, "chr1p"],
            [pd.NA, pd.NA, pd.NA, "chr1q"],
            ["chrX", 1, 8, "chrX_0"],
        ],
        columns=colnames,
    ).astype({"start": pd.Int64Dtype(), "end": pd.Int64Dtype()})
    df_result = pd.DataFrame(
        [["chr1", -6, 12, "chr1p"]],
        columns=colnames,
    ).astype({"start": pd.Int64Dtype(), "end": pd.Int64Dtype()})

    region1 = "chr1:0-1"
    pd.testing.assert_frame_equal(
        df_result, bioframe.select(df, region1).reset_index(drop=True)
    )


def test_trim():

    ### trim with view_df
    view_df = pd.DataFrame(
        [
            ["chr1", 0, 12, "chr1p"],
            ["chr1", 13, 26, "chr1q"],
            ["chrX", 1, 8, "chrX_0"],
        ],
        columns=["chrom", "start", "end", "name"],
    )
    df = pd.DataFrame(
        [
            ["chr1", -6, 12, "chr1p"],
            ["chr1", 0, 12, "chr1p"],
            ["chr1", 32, 36, "chr1q"],
            ["chrX", 1, 8, "chrX_0"],
        ],
        columns=["chrom", "start", "end", "view_region"],
    )
    df_trimmed = pd.DataFrame(
        [
            ["chr1", 0, 12, "chr1p"],
            ["chr1", 0, 12, "chr1p"],
            ["chr1", 26, 26, "chr1q"],
            ["chrX", 1, 8, "chrX_0"],
        ],
        columns=["chrom", "start", "end", "view_region"],
    )
    with pytest.raises(ValueError):
        bioframe.trim(df, view_df=view_df)

    # df_view_col already exists, so need to specify it:
    pd.testing.assert_frame_equal(
        df_trimmed, bioframe.trim(df, view_df=view_df, df_view_col="view_region")
    )

    # non-default columns in view
    funky_view = view_df.rename(columns={"chrom": "chr"})
    pd.testing.assert_frame_equal(
        df_trimmed,
        bioframe.trim(
            df,
            view_df=funky_view,
            df_view_col="view_region",
            cols_view=["chr", "start", "end"],
        ),
    )

    ### trim with view_df interpreted from dictionary for chromsizes
    chromsizes = {"chr1": 20, "chrX_0": 5}
    df = pd.DataFrame(
        [
            ["chr1", 0, 12],
            ["chr1", 13, 26],
            ["chrX_0", 1, 8],
        ],
        columns=["chrom", "startFunky", "end"],
    )
    df_trimmed = pd.DataFrame(
        [
            ["chr1", 0, 12],
            ["chr1", 13, 20],
            ["chrX_0", 1, 5],
        ],
        columns=["chrom", "startFunky", "end"],
    ).astype({"startFunky": pd.Int64Dtype(), "end": pd.Int64Dtype()})
    pd.testing.assert_frame_equal(
        df_trimmed,
        bioframe.trim(
            df,
            view_df=chromsizes,
            cols=["chrom", "startFunky", "end"],
            return_view_columns=False,
        ),
    )

    ### trim with default limits=None and negative values
    df = pd.DataFrame(
        [
            ["chr1", -4, 12],
            ["chr1", 13, 26],
            ["chrX", -5, -1],
        ],
        columns=["chrom", "start", "end"],
    )
    df_trimmed = pd.DataFrame(
        [
            ["chr1", 0, 12],
            ["chr1", 13, 26],
            ["chrX", 0, 0],
        ],
        columns=["chrom", "start", "end"],
    )
    pd.testing.assert_frame_equal(df_trimmed, bioframe.trim(df))

    ### trim when there are NaN intervals
    df = pd.DataFrame(
        [
            ["chr1", -4, 12, "chr1p"],
            [pd.NA, pd.NA, pd.NA, "chr1q"],
            ["chrX", -5, -1, "chrX_0"],
        ],
        columns=["chrom", "start", "end", "region"],
    ).astype({"start": pd.Int64Dtype(), "end": pd.Int64Dtype()})
    df_trimmed = pd.DataFrame(
        [
            ["chr1", 0, 12, "chr1p"],
            [pd.NA, pd.NA, pd.NA, "chr1q"],
            ["chrX", 0, 0, "chrX_0"],
        ],
        columns=["chrom", "start", "end", "region"],
    ).astype({"start": pd.Int64Dtype(), "end": pd.Int64Dtype()})
    pd.testing.assert_frame_equal(df_trimmed, bioframe.trim(df))

    ### trim with view_df and NA intervals
    view_df = pd.DataFrame(
        [
            ["chr1", 0, 12, "chr1p"],
            ["chr1", 13, 26, "chr1q"],
            ["chrX", 1, 12, "chrX_0"],
        ],
        columns=["chrom", "start", "end", "name"],
    )
    df = pd.DataFrame(
        [
            ["chr1", -6, 12],
            ["chr1", 0, 12],
            [pd.NA, pd.NA, pd.NA],
            ["chrX", 1, 20],
        ],
        columns=["chrom", "start", "end"],
    ).astype({"start": pd.Int64Dtype(), "end": pd.Int64Dtype()})
    df_trimmed = pd.DataFrame(
        [
            ["chr1", 0, 12, "chr1p"],
            ["chr1", 0, 12, "chr1p"],
            [pd.NA, pd.NA, pd.NA, pd.NA],
            ["chrX", 1, 12, "chrX_0"],
        ],
        columns=["chrom", "start", "end", "view_region"],
    ).astype({"start": pd.Int64Dtype(), "end": pd.Int64Dtype()})
    # infer df_view_col with assign_view and ignore NAs
    pd.testing.assert_frame_equal(
        df_trimmed,
        bioframe.trim(df, view_df=view_df, df_view_col=None, return_view_columns=True)[
            ["chrom", "start", "end", "view_region"]
        ],
    )


def test_expand():

    d = """chrom  start  end
         0  chr1      1    5
         1  chr1     50   55
         2  chr2    100  200"""
    fake_bioframe = pd.read_csv(StringIO(d), sep=r"\s+")

    expand_bp = 10
    fake_expanded = bioframe.expand(fake_bioframe, expand_bp)
    d = """chrom  start  end
         0  chr1      -9    15
         1  chr1     40   65
         2  chr2    90  210"""
    df = pd.read_csv(StringIO(d), sep=r"\s+")
    pd.testing.assert_frame_equal(df, fake_expanded)

    # expand with negative pad
    expand_bp = -10
    fake_expanded = bioframe.expand(fake_bioframe, expand_bp)
    d = """chrom  start  end
         0  chr1      3    3
         1  chr1     52   52
         2  chr2    110  190"""
    df = pd.read_csv(StringIO(d), sep=r"\s+")
    pd.testing.assert_frame_equal(df, fake_expanded)

    expand_bp = -10
    fake_expanded = bioframe.expand(fake_bioframe, expand_bp, side="left")
    d = """chrom  start  end
         0  chr1      3    5
         1  chr1     52   55
         2  chr2    110  200"""
    df = pd.read_csv(StringIO(d), sep=r"\s+")
    pd.testing.assert_frame_equal(df, fake_expanded)

    # expand with multiplicative pad
    mult = 0
    fake_expanded = bioframe.expand(fake_bioframe, pad=None, scale=mult)
    d = """chrom  start  end
         0  chr1      3    3
         1  chr1     52   52
         2  chr2    150  150"""
    df = pd.read_csv(StringIO(d), sep=r"\s+")
    pd.testing.assert_frame_equal(df, fake_expanded)

    mult = 2.0
    fake_expanded = bioframe.expand(fake_bioframe, pad=None, scale=mult)
    d = """chrom  start  end
         0  chr1      -1    7
         1  chr1     48   58
         2  chr2    50  250"""
    df = pd.read_csv(StringIO(d), sep=r"\s+")
    pd.testing.assert_frame_equal(df, fake_expanded)

    # expand with NA and non-integer multiplicative pad
    d = """chrom  start  end
         0  chr1      1    5
         1  NA     NA   NA
         2  chr2    100  200"""
    fake_bioframe = pd.read_csv(StringIO(d), sep=r"\s+").astype(
        {"start": pd.Int64Dtype(), "end": pd.Int64Dtype()}
    )
    mult = 1.10
    fake_expanded = bioframe.expand(fake_bioframe, pad=None, scale=mult)
    d = """chrom  start  end
         0  chr1      1   5
         1  NA     NA   NA
         2  chr2    95  205"""
    df = pd.read_csv(StringIO(d), sep=r"\s+").astype(
        {"start": pd.Int64Dtype(), "end": pd.Int64Dtype()}
    )
    pd.testing.assert_frame_equal(df, fake_expanded)


def test_overlap():

    ### test consistency of overlap(how='inner') with pyranges.join ###
    ### note does not test overlap_start or overlap_end columns of bioframe.overlap
    df1 = mock_bioframe()
    df2 = mock_bioframe()
    assert df1.equals(df2) == False

    # p1 = bioframe_to_pyranges(df1)
    # p2 = bioframe_to_pyranges(df2)
    # pp = pyranges_overlap_to_bioframe(p1.join(p2, how=None))[
    #     ["chrom_1", "start_1", "end_1", "chrom_2", "start_2", "end_2"]
    # ]
    # bb = bioframe.overlap(df1, df2, how="inner")[
    #     ["chrom_1", "start_1", "end_1", "chrom_2", "start_2", "end_2"]
    # ]
    # pp = pp.sort_values(
    #     ["chrom_1", "start_1", "end_1", "chrom_2", "start_2", "end_2"],
    #     ignore_index=True,
    # )
    # bb = bb.sort_values(
    #     ["chrom_1", "start_1", "end_1", "chrom_2", "start_2", "end_2"],
    #     ignore_index=True,
    # )
    # pd.testing.assert_frame_equal(bb, pp, check_dtype=False, check_exact=False)
    # print("overlap elements agree")

    ### test overlap on= [] ###
    df1 = pd.DataFrame(
        [
            ["chr1", 8, 12, "+", "cat"],
            ["chr1", 8, 12, "-", "cat"],
            ["chrX", 1, 8, "+", "cat"],
        ],
        columns=["chrom1", "start", "end", "strand", "animal"],
    )

    df2 = pd.DataFrame(
        [["chr1", 6, 10, "+", "dog"], ["chrX", 7, 10, "-", "dog"]],
        columns=["chrom2", "start2", "end2", "strand", "animal"],
    )

    b = bioframe.overlap(
        df1,
        df2,
        on=["animal"],
        how="left",
        cols1=("chrom1", "start", "end"),
        cols2=("chrom2", "start2", "end2"),
        return_index=True,
        return_input=False,
    )
    assert np.sum(pd.isna(b["index_"].values)) == 3

    b = bioframe.overlap(
        df1,
        df2,
        on=["strand"],
        how="left",
        cols1=("chrom1", "start", "end"),
        cols2=("chrom2", "start2", "end2"),
        return_index=True,
        return_input=False,
    )
    assert np.sum(pd.isna(b["index_"].values)) == 2

    b = bioframe.overlap(
        df1,
        df2,
        on=None,
        how="left",
        cols1=("chrom1", "start", "end"),
        cols2=("chrom2", "start2", "end2"),
        return_index=True,
        return_input=False,
    )
    assert np.sum(pd.isna(b["index_"].values)) == 0

    ### test overlap 'left', 'outer', and 'right'
    b = bioframe.overlap(
        df1,
        df2,
        on=None,
        how="outer",
        cols1=("chrom1", "start", "end"),
        cols2=("chrom2", "start2", "end2"),
    )
    assert len(b) == 3

    b = bioframe.overlap(
        df1,
        df2,
        on=["animal"],
        how="outer",
        cols1=("chrom1", "start", "end"),
        cols2=("chrom2", "start2", "end2"),
    )
    assert len(b) == 5

    b = bioframe.overlap(
        df1,
        df2,
        on=["animal"],
        how="inner",
        cols1=("chrom1", "start", "end"),
        cols2=("chrom2", "start2", "end2"),
    )
    assert len(b) == 0

    b = bioframe.overlap(
        df1,
        df2,
        on=["animal"],
        how="right",
        cols1=("chrom1", "start", "end"),
        cols2=("chrom2", "start2", "end2"),
    )
    assert len(b) == 2

    b = bioframe.overlap(
        df1,
        df2,
        on=["animal"],
        how="left",
        cols1=("chrom1", "start", "end"),
        cols2=("chrom2", "start2", "end2"),
    )
    assert len(b) == 3

    ### test keep_order and NA handling
    df1 = pd.DataFrame(
        [
            ["chr1", 8, 12, "+"],
            [pd.NA, pd.NA, pd.NA, "-"],
            ["chrX", 1, 8, "+"],
        ],
        columns=["chrom", "start", "end", "strand"],
    ).astype({"start": pd.Int64Dtype(), "end": pd.Int64Dtype()})

    df2 = pd.DataFrame(
        [["chr1", 6, 10, "+"], [pd.NA, pd.NA, pd.NA, "-"], ["chrX", 7, 10, "-"]],
        columns=["chrom2", "start2", "end2", "strand"],
    ).astype({"start2": pd.Int64Dtype(), "end2": pd.Int64Dtype()})

    assert df1.equals(
        bioframe.overlap(
            df1, df2, how="left", keep_order=True, cols2=["chrom2", "start2", "end2"]
        )[["chrom", "start", "end", "strand"]]
    )
    assert ~df1.equals(
        bioframe.overlap(
            df1, df2, how="left", keep_order=False, cols2=["chrom2", "start2", "end2"]
        )[["chrom", "start", "end", "strand"]]
    )

    df1 = pd.DataFrame(
        [
            ["chr1", 8, 12, "+", pd.NA],
            [pd.NA, pd.NA, pd.NA, "-", pd.NA],
            ["chrX", 1, 8, pd.NA, pd.NA],
        ],
        columns=["chrom", "start", "end", "strand", "animal"],
    ).astype({"start": pd.Int64Dtype(), "end": pd.Int64Dtype()})

    df2 = pd.DataFrame(
        [["chr1", 6, 10, pd.NA, "tiger"]],
        columns=["chrom2", "start2", "end2", "strand", "animal"],
    ).astype({"start2": pd.Int64Dtype(), "end2": pd.Int64Dtype()})

    assert (
        bioframe.overlap(
            df1,
            df2,
            how="outer",
            cols2=["chrom2", "start2", "end2"],
            return_index=True,
            keep_order=False,
        ).shape
        == (3, 12)
    )

    ### result of overlap should still have bedframe-like properties
    overlap_df = bioframe.overlap(
        df1,
        df2,
        how="outer",
        cols2=["chrom2", "start2", "end2"],
        return_index=True,
        suffixes=("", ""),
    )
    assert checks.is_bedframe(
        overlap_df[df1.columns],
    )
    assert checks.is_bedframe(
        overlap_df[df2.columns], cols=["chrom2", "start2", "end2"]
    )

    overlap_df = bioframe.overlap(
        df1,
        df2,
        how="innter",
        cols2=["chrom2", "start2", "end2"],
        return_index=True,
        suffixes=("", ""),
    )
    assert checks.is_bedframe(
        overlap_df[df1.columns],
    )
    assert checks.is_bedframe(
        overlap_df[df2.columns], cols=["chrom2", "start2", "end2"]
    )

    # test keep_order incompatible if how!= 'left'
    with pytest.raises(ValueError):
        bioframe.overlap(
            df1,
            df2,
            how="outer",
            on=["animal"],
            cols2=["chrom2", "start2", "end2"],
            keep_order=True,
        )


def test_cluster():
    df1 = pd.DataFrame(
        [
            ["chr1", 1, 5],
            ["chr1", 3, 8],
            ["chr1", 8, 10],
            ["chr1", 12, 14],
        ],
        columns=["chrom", "start", "end"],
    )
    df_annotated = bioframe.cluster(df1)
    assert (
        df_annotated["cluster"].values == np.array([0, 0, 0, 1])
    ).all()  # the last interval does not overlap the first three
    df_annotated = bioframe.cluster(df1, min_dist=2)
    assert (
        df_annotated["cluster"].values == np.array([0, 0, 0, 0])
    ).all()  # all intervals part of the same cluster

    df_annotated = bioframe.cluster(df1, min_dist=None)
    assert (
        df_annotated["cluster"].values == np.array([0, 0, 1, 2])
    ).all()  # adjacent intervals not clustered

    df1.iloc[0, 0] = "chrX"
    df_annotated = bioframe.cluster(df1)
    assert (
        df_annotated["cluster"].values == np.array([2, 0, 0, 1])
    ).all()  # do not cluster intervals across chromosomes

    # test consistency with pyranges (which automatically sorts df upon creation and uses 1-based indexing for clusters)
    # assert (
    #     (bioframe_to_pyranges(df1).cluster(count=True).df["Cluster"].values - 1)
    #     == bioframe.cluster(df1.sort_values(["chrom", "start"]))["cluster"].values
    # ).all()

    # test on=[] argument
    df1 = pd.DataFrame(
        [
            ["chr1", 3, 8, "+", "cat", 5.5],
            ["chr1", 3, 8, "-", "dog", 6.5],
            ["chr1", 6, 10, "-", "cat", 6.5],
            ["chrX", 6, 10, "-", "cat", 6.5],
        ],
        columns=["chrom", "start", "end", "strand", "animal", "location"],
    )
    assert (
        bioframe.cluster(df1, on=["animal"])["cluster"].values == np.array([0, 1, 0, 2])
    ).all()
    assert (
        bioframe.cluster(df1, on=["strand"])["cluster"].values == np.array([0, 1, 1, 2])
    ).all()
    assert (
        bioframe.cluster(df1, on=["location", "animal"])["cluster"].values
        == np.array([0, 2, 1, 3])
    ).all()

    ### test cluster with NAs
    df1 = pd.DataFrame(
        [
            ["chrX", 1, 8, pd.NA, pd.NA],
            [pd.NA, pd.NA, pd.NA, "-", pd.NA],
            ["chr1", 8, 12, "+", pd.NA],
            ["chr1", 1, 8, np.nan, pd.NA],
            [pd.NA, np.nan, pd.NA, "-", pd.NA],
        ],
        columns=["chrom", "start", "end", "strand", "animal"],
    ).astype({"start": pd.Int64Dtype(), "end": pd.Int64Dtype()})

    assert bioframe.cluster(df1)["cluster"].max() == 3
    assert bioframe.cluster(df1, on=["strand"])["cluster"].max() == 4
    pd.testing.assert_frame_equal(df1, bioframe.cluster(df1)[df1.columns])

    assert checks.is_bedframe(
        bioframe.cluster(df1, on=["strand"]),
        cols=["chrom", "cluster_start", "cluster_end"],
    )
    assert checks.is_bedframe(
        bioframe.cluster(df1), cols=["chrom", "cluster_start", "cluster_end"]
    )
    assert checks.is_bedframe(bioframe.cluster(df1))


def test_merge():
    df1 = pd.DataFrame(
        [
            ["chr1", 1, 5],
            ["chr1", 3, 8],
            ["chr1", 8, 10],
            ["chr1", 12, 14],
        ],
        columns=["chrom", "start", "end"],
    )

    # the last interval does not overlap the first three with default min_dist=0
    assert (bioframe.merge(df1)["n_intervals"].values == np.array([3, 1])).all()

    # adjacent intervals are not clustered with min_dist=none
    assert (
        bioframe.merge(df1, min_dist=None)["n_intervals"].values == np.array([2, 1, 1])
    ).all()

    # all intervals part of one cluster
    assert (
        bioframe.merge(df1, min_dist=2)["n_intervals"].values == np.array([4])
    ).all()

    df1.iloc[0, 0] = "chrX"
    assert (
        bioframe.merge(df1, min_dist=None)["n_intervals"].values
        == np.array([1, 1, 1, 1])
    ).all()
    assert (
        bioframe.merge(df1, min_dist=0)["n_intervals"].values == np.array([2, 1, 1])
    ).all()

    # total number of intervals should equal length of original dataframe
    mock_df = mock_bioframe()
    assert np.sum(bioframe.merge(mock_df, min_dist=0)["n_intervals"].values) == len(
        mock_df
    )

    # # test consistency with pyranges
    # pd.testing.assert_frame_equal(
    #     pyranges_to_bioframe(bioframe_to_pyranges(df1).merge(count=True)),
    #     bioframe.merge(df1),
    #     check_dtype=False,
    #     check_exact=False,
    # )

    # test on=['chrom',...] argument
    df1 = pd.DataFrame(
        [
            ["chr1", 3, 8, "+", "cat", 5.5],
            ["chr1", 3, 8, "-", "dog", 6.5],
            ["chr1", 6, 10, "-", "cat", 6.5],
            ["chrX", 6, 10, "-", "cat", 6.5],
        ],
        columns=["chrom", "start", "end", "strand", "animal", "location"],
    )
    assert len(bioframe.merge(df1, on=None)) == 2
    assert len(bioframe.merge(df1, on=["strand"])) == 3
    assert len(bioframe.merge(df1, on=["strand", "location"])) == 3
    assert len(bioframe.merge(df1, on=["strand", "location", "animal"])) == 4
    d = """ chrom   start   end animal  n_intervals
        0   chr1    3   10  cat 2
        1   chr1    3   8   dog 1
        2   chrX    6   10  cat 1"""
    df = pd.read_csv(StringIO(d), sep=r"\s+")
    pd.testing.assert_frame_equal(
        df,
        bioframe.merge(df1, on=["animal"]),
        check_dtype=False,
    )

    # merge with repeated indices
    df = pd.DataFrame(
        {"chrom": ["chr1", "chr2"], "start": [100, 400], "end": [110, 410]}
    )
    df.index = [0, 0]
    pd.testing.assert_frame_equal(
        df.reset_index(drop=True), bioframe.merge(df)[["chrom", "start", "end"]]
    )

    # test merge with NAs
    df1 = pd.DataFrame(
        [
            ["chrX", 1, 8, pd.NA, pd.NA],
            [pd.NA, pd.NA, pd.NA, "-", pd.NA],
            ["chr1", 8, 12, "+", pd.NA],
            ["chr1", 1, 8, np.nan, pd.NA],
            [pd.NA, np.nan, pd.NA, "-", pd.NA],
        ],
        columns=["chrom", "start", "end", "strand", "animal"],
    ).astype({"start": pd.Int64Dtype(), "end": pd.Int64Dtype()})

    assert bioframe.merge(df1).shape[0] == 4
    assert bioframe.merge(df1)["start"].iloc[0] == 1
    assert bioframe.merge(df1)["end"].iloc[0] == 12
    assert bioframe.merge(df1, on=["strand"]).shape[0] == df1.shape[0]
    assert bioframe.merge(df1, on=["animal"]).shape[0] == df1.shape[0]
    assert bioframe.merge(df1, on=["animal"]).shape[1] == df1.shape[1] + 1
    assert checks.is_bedframe(bioframe.merge(df1, on=["strand", "animal"]))


def test_complement():
    ### complementing a df with no intervals in chrX by a view with chrX should return entire chrX region
    df1 = pd.DataFrame(
        [["chr1", 1, 5], ["chr1", 3, 8], ["chr1", 8, 10], ["chr1", 12, 14]],
        columns=["chrom", "start", "end"],
    )
    df1_chromsizes = {"chr1": 100, "chrX": 100}

    df1_complement = pd.DataFrame(
        [
            ["chr1", 0, 1, "chr1"],
            ["chr1", 10, 12, "chr1"],
            ["chr1", 14, 100, "chr1"],
            ["chrX", 0, 100, "chrX"],
        ],
        columns=["chrom", "start", "end", "view_region"],
    )

    pd.testing.assert_frame_equal(
        bioframe.complement(df1, view_df=df1_chromsizes), df1_complement
    )

    # non-default columns in view
    funky_view = make_viewframe(df1_chromsizes, cols=["chr", "start", "end"])

    pd.testing.assert_frame_equal(
        bioframe.complement(
            df1,
            view_df=funky_view,
            cols_view=["chr", "start", "end"],
        ),
        df1_complement,
    )

    ### test complement with two chromosomes ###
    df1.iloc[0, 0] = "chrX"
    df1_complement = pd.DataFrame(
        [
            ["chr1", 0, 3, "chr1"],
            ["chr1", 10, 12, "chr1"],
            ["chr1", 14, 100, "chr1"],
            ["chrX", 0, 1, "chrX"],
            ["chrX", 5, 100, "chrX"],
        ],
        columns=["chrom", "start", "end", "view_region"],
    )
    pd.testing.assert_frame_equal(
        bioframe.complement(df1, view_df=df1_chromsizes), df1_complement
    )

    ### test complement with no view_df and a negative interval
    df1 = pd.DataFrame(
        [["chr1", -5, 5], ["chr1", 10, 20]], columns=["chrom", "start", "end"]
    )
    df1_complement = pd.DataFrame(
        [
            ["chr1", 5, 10, "chr1"],
            ["chr1", 20, np.iinfo(np.int64).max, "chr1"],
        ],
        columns=["chrom", "start", "end", "view_region"],
    )
    pd.testing.assert_frame_equal(bioframe.complement(df1), df1_complement)

    ### test complement with an overhanging interval
    df1 = pd.DataFrame(
        [["chr1", -5, 5], ["chr1", 10, 20]], columns=["chrom", "start", "end"]
    )
    chromsizes = bioframe.make_viewframe(
        {"chr1": 15}, name_style="ucsc", view_name_col="VR"
    )
    df1_complement = pd.DataFrame(
        [
            ["chr1", 5, 10, "chr1:0-15"],
        ],
        columns=["chrom", "start", "end", "view_region"],
    )
    pd.testing.assert_frame_equal(
        bioframe.complement(df1, view_df=chromsizes, view_name_col="VR"), df1_complement
    )

    ### test complement where an interval from df overlaps two different regions from view
    ### test complement with no view_df and a negative interval
    df1 = pd.DataFrame([["chr1", 5, 15]], columns=["chrom", "start", "end"])
    chromsizes = [("chr1", 0, 9, "chr1p"), ("chr1", 11, 20, "chr1q")]
    df1_complement = pd.DataFrame(
        [["chr1", 0, 5, "chr1p"], ["chr1", 15, 20, "chr1q"]],
        columns=["chrom", "start", "end", "view_region"],
    )
    pd.testing.assert_frame_equal(bioframe.complement(df1, chromsizes), df1_complement)

    ### test complement with NAs
    df1 = pd.DataFrame(
        [[pd.NA, pd.NA, pd.NA], ["chr1", 5, 15], [pd.NA, pd.NA, pd.NA]],
        columns=["chrom", "start", "end"],
    ).astype(
        {
            "start": pd.Int64Dtype(),
            "end": pd.Int64Dtype(),
        }
    )

    pd.testing.assert_frame_equal(bioframe.complement(df1, chromsizes), df1_complement)

    with pytest.raises(ValueError):  # no NAs allowed in chromsizes
        bioframe.complement(
            df1, [("chr1", pd.NA, 9, "chr1p"), ("chr1", 11, 20, "chr1q")]
        )

    assert checks.is_bedframe(bioframe.complement(df1, chromsizes))


def test_closest():
    df1 = pd.DataFrame(
        [
            ["chr1", 1, 5],
        ],
        columns=["chrom", "start", "end"],
    )

    df2 = pd.DataFrame(
        [["chr1", 4, 8], ["chr1", 10, 11]], columns=["chrom", "start", "end"]
    )

    ### closest(df1,df2,k=1) ###
    d = """chrom  start  end chrom_  start_  end_  distance
        0    chr1        1      5    chr1        4      8         0"""
    df = pd.read_csv(StringIO(d), sep=r"\s+").astype(
        {
            "start_": pd.Int64Dtype(),
            "end_": pd.Int64Dtype(),
            "distance": pd.Int64Dtype(),
        }
    )
    pd.testing.assert_frame_equal(df, bioframe.closest(df1, df2, k=1))

    ### closest(df1,df2, ignore_overlaps=True)) ###
    d = """chrom_1 start_1 end_1   chrom_2 start_2 end_2   distance
        0   chr1    1   5   chr1    10  11  5"""
    df = pd.read_csv(StringIO(d), sep=r"\s+").astype(
        {
            "start_2": pd.Int64Dtype(),
            "end_2": pd.Int64Dtype(),
            "distance": pd.Int64Dtype(),
        }
    )
    pd.testing.assert_frame_equal(
        df, bioframe.closest(df1, df2, suffixes=("_1", "_2"), ignore_overlaps=True)
    )

    ### closest(df1,df2,k=2) ###
    d = """chrom_1 start_1 end_1   chrom_2 start_2 end_2   distance
            0   chr1    1   5   chr1    4   8   0
            1   chr1    1   5   chr1    10  11  5"""
    df = pd.read_csv(StringIO(d), sep=r"\s+").astype(
        {
            "start_2": pd.Int64Dtype(),
            "end_2": pd.Int64Dtype(),
            "distance": pd.Int64Dtype(),
        }
    )
    pd.testing.assert_frame_equal(
        df, bioframe.closest(df1, df2, suffixes=("_1", "_2"), k=2)
    )

    ### closest(df2,df1) ###
    d = """chrom_1  start_1 end_1   chrom_2 start_2 end_2   distance
            0   chr1    4   8   chr1    1   5   0
            1   chr1    10  11  chr1    1   5   5 """
    df = pd.read_csv(StringIO(d), sep=r"\s+").astype(
        {
            "start_2": pd.Int64Dtype(),
            "end_2": pd.Int64Dtype(),
            "distance": pd.Int64Dtype(),
        }
    )
    pd.testing.assert_frame_equal(df, bioframe.closest(df2, df1, suffixes=("_1", "_2")))

    ### change first interval to new chrom ###
    df2.iloc[0, 0] = "chrA"
    d = """chrom start   end     chrom_ start_ end_  distance
              0   chr1    1   5   chr1    10  11  5"""
    df = pd.read_csv(StringIO(d), sep=r"\s+").astype(
        {
            "start_": pd.Int64Dtype(),
            "end_": pd.Int64Dtype(),
            "distance": pd.Int64Dtype(),
        }
    )
    pd.testing.assert_frame_equal(df, bioframe.closest(df1, df2, k=1))

    ### test other return arguments ###
    df2.iloc[0, 0] = "chr1"
    d = """
        index index_ have_overlap    overlap_start   overlap_end distance
        0   0   0   True    4   5   0
        1   0   1   False   <NA>    <NA>    5
        """
    df = pd.read_csv(StringIO(d), sep=r"\s+")
    pd.testing.assert_frame_equal(
        df,
        bioframe.closest(
            df1,
            df2,
            k=2,
            return_overlap=True,
            return_index=True,
            return_input=False,
            return_distance=True,
        ),
        check_dtype=False,
    )

    # closest should ignore empty groups (e.g. from categorical chrom)
    df = pd.DataFrame(
        [
            ["chrX", 1, 8],
            ["chrX", 2, 10],
        ],
        columns=["chrom", "start", "end"],
    )
    d = """ chrom_1 start_1 end_1   chrom_2 start_2 end_2   distance
            0   chrX    1   8   chrX    2   10  0
            1   chrX    2   10  chrX    1   8   0"""
    df_closest = pd.read_csv(StringIO(d), sep=r"\s+")
    df_cat = pd.CategoricalDtype(categories=["chrX", "chr1"], ordered=True)
    df = df.astype({"chrom": df_cat})
    pd.testing.assert_frame_equal(
        df_closest,
        bioframe.closest(df, suffixes=("_1", "_2")),
        check_dtype=False,
        check_categorical=False,
    )

    # closest should ignore null rows: code will need to be modified
    # as for overlap if an on=[] option is added
    df1 = pd.DataFrame(
        [
            [pd.NA, pd.NA, pd.NA],
            ["chr1", 1, 5],
        ],
        columns=["chrom", "start", "end"],
    ).astype({"start": pd.Int64Dtype(), "end": pd.Int64Dtype()})

    df2 = pd.DataFrame(
        [
            [pd.NA, pd.NA, pd.NA],
            ["chr1", 4, 8],
            [pd.NA, pd.NA, pd.NA],
            ["chr1", 10, 11],
        ],
        columns=["chrom", "start", "end"],
    ).astype({"start": pd.Int64Dtype(), "end": pd.Int64Dtype()})

    d = """chrom_1 start_1 end_1   chrom_2 start_2 end_2   distance
        0   chr1    1   5   chr1    10  11  5"""
    df = pd.read_csv(StringIO(d), sep=r"\s+").astype(
        {
            "start_1": pd.Int64Dtype(),
            "end_1": pd.Int64Dtype(),
            "start_2": pd.Int64Dtype(),
            "end_2": pd.Int64Dtype(),
            "distance": pd.Int64Dtype(),
        }
    )
    pd.testing.assert_frame_equal(
        df, bioframe.closest(df1, df2, suffixes=("_1", "_2"), ignore_overlaps=True, k=5)
    )

    with pytest.raises(ValueError):  # inputs must be valid bedFrames
        df1.iloc[0, 0] = "chr10"
        bioframe.closest(df1, df2)


def test_coverage():

    #### coverage does not exceed length of original interval
    df1 = pd.DataFrame([["chr1", 3, 8]], columns=["chrom", "start", "end"])
    df2 = pd.DataFrame([["chr1", 2, 10]], columns=["chrom", "start", "end"])
    d = """chrom    start   end coverage
         0  chr1    3   8   5"""
    df = pd.read_csv(StringIO(d), sep=r"\s+")
    pd.testing.assert_frame_equal(df, bioframe.coverage(df1, df2))

    ### coverage of interval on different chrom returns zero for coverage and n_overlaps
    df1 = pd.DataFrame([["chr1", 3, 8]], columns=["chrom", "start", "end"])
    df2 = pd.DataFrame([["chrX", 3, 8]], columns=["chrom", "start", "end"])
    d = """chrom    start   end coverage
        0  chr1      3       8     0   """
    df = pd.read_csv(StringIO(d), sep=r"\s+")
    pd.testing.assert_frame_equal(df, bioframe.coverage(df1, df2))

    ### when a second overlap starts within the first
    df1 = pd.DataFrame([["chr1", 3, 8]], columns=["chrom", "start", "end"])
    df2 = pd.DataFrame(
        [["chr1", 3, 6], ["chr1", 5, 8]], columns=["chrom", "start", "end"]
    )
    d = """chrom    start   end coverage
         0  chr1     3       8     5"""
    df = pd.read_csv(StringIO(d), sep=r"\s+")
    pd.testing.assert_frame_equal(df, bioframe.coverage(df1, df2))

    ### coverage of NA interval returns zero for coverage
    df1 = pd.DataFrame(
        [
            ["chr1", 10, 20],
            [pd.NA, pd.NA, pd.NA],
            ["chr1", 3, 8],
            [pd.NA, pd.NA, pd.NA],
        ],
        columns=["chrom", "start", "end"],
    )
    df2 = pd.DataFrame(
        [["chr1", 3, 6], ["chr1", 5, 8], [pd.NA, pd.NA, pd.NA]],
        columns=["chrom", "start", "end"],
    )
    df1 = bioframe.sanitize_bedframe(df1)
    df2 = bioframe.sanitize_bedframe(df2)

    df_coverage = pd.DataFrame(
        [
            ["chr1", 10, 20, 0],
            [pd.NA, pd.NA, pd.NA, 0],
            ["chr1", 3, 8, 5],
            [pd.NA, pd.NA, pd.NA, 0],
        ],
        columns=["chrom", "start", "end", "coverage"],
    ).astype(
        {"start": pd.Int64Dtype(), "end": pd.Int64Dtype(), "coverage": pd.Int64Dtype()}
    )
    pd.testing.assert_frame_equal(df_coverage, bioframe.coverage(df1, df2))

    ### coverage without return_input returns a single column dataFrame
    assert (
        bioframe.coverage(df1, df2, return_input=False)["coverage"].values
        == np.array([0, 0, 5, 0])
    ).all()


def test_subtract():
    ### no intervals should be left after self-subtraction
    df1 = pd.DataFrame(
        [["chrX", 3, 8], ["chr1", 4, 7], ["chrX", 1, 5]],
        columns=["chrom", "start", "end"],
    )
    assert len(bioframe.subtract(df1, df1)) == 0

    ### no intervals on chrX should remain after subtracting a longer interval
    ### interval on chr1 should be split.
    ### additional column should be propagated to children.
    df2 = pd.DataFrame(
        [
            ["chrX", 0, 18],
            ["chr1", 5, 6],
        ],
        columns=["chrom", "start", "end"],
    )
    df1["animal"] = "sea-creature"
    df_result = pd.DataFrame(
        [["chr1", 4, 5, "sea-creature"], ["chr1", 6, 7, "sea-creature"]],
        columns=["chrom", "start", "end", "animal"],
    ).astype({"start": pd.Int64Dtype(), "end": pd.Int64Dtype()})
    pd.testing.assert_frame_equal(
        df_result,
        bioframe.subtract(df1, df2)
        .sort_values(["chrom", "start", "end"])
        .reset_index(drop=True),
    )

    ### no intervals on chrX should remain after subtracting a longer interval
    df2 = pd.DataFrame(
        [["chrX", 0, 4], ["chr1", 6, 6], ["chrX", 4, 9]],
        columns=["chrom", "start", "end"],
    )

    df1["animal"] = "sea-creature"
    df_result = pd.DataFrame(
        [["chr1", 4, 6, "sea-creature"], ["chr1", 6, 7, "sea-creature"]],
        columns=["chrom", "start", "end", "animal"],
    )
    pd.testing.assert_frame_equal(
        df_result.astype({"start": pd.Int64Dtype(), "end": pd.Int64Dtype()}),
        bioframe.subtract(df1, df2)
        .sort_values(["chrom", "start", "end"])
        .reset_index(drop=True),
    )

    ### subtracting dataframes funny column names

    funny_cols = ["C", "chromStart", "chromStop"]

    df1 = pd.DataFrame(
        [["chrX", 3, 8], ["chr1", 4, 7], ["chrX", 1, 5]],
        columns=funny_cols,
    )
    df1["strand"] = "+"

    assert len(bioframe.subtract(df1, df1, cols1=funny_cols, cols2=funny_cols)) == 0

    funny_cols2 = ["chr", "st", "e"]
    df2 = pd.DataFrame(
        [
            ["chrX", 0, 18],
            ["chr1", 5, 6],
        ],
        columns=funny_cols2,
    )

    df_result = pd.DataFrame(
        [["chr1", 4, 5, "+"], ["chr1", 6, 7, "+"]],
        columns=funny_cols + ["strand"],
    )
    df_result = df_result.astype(
        {funny_cols[1]: pd.Int64Dtype(), funny_cols[2]: pd.Int64Dtype()}
    )

    pd.testing.assert_frame_equal(
        df_result,
        bioframe.subtract(df1, df2, cols1=funny_cols, cols2=funny_cols2)
        .sort_values(funny_cols)
        .reset_index(drop=True),
    )

    # subtract should ignore empty groups
    df1 = pd.DataFrame(
        [
            ["chrX", 1, 8],
            ["chrX", 2, 10],
        ],
        columns=["chrom", "start", "end"],
    )
    df2 = pd.DataFrame(
        [
            ["chrX", 1, 8],
        ],
        columns=["chrom", "start", "end"],
    )
    df_cat = pd.CategoricalDtype(categories=["chrX", "chr1"], ordered=True)
    df1 = df1.astype({"chrom": df_cat})
    df_subtracted = pd.DataFrame(
        [
            ["chrX", 8, 10],
        ],
        columns=["chrom", "start", "end"],
    )

    assert bioframe.subtract(df1, df1).empty

    pd.testing.assert_frame_equal(
        df_subtracted.astype({"start": pd.Int64Dtype(), "end": pd.Int64Dtype()}),
        bioframe.subtract(df1, df2),
        check_dtype=False,
        check_categorical=False,
    )

    ## test transferred from deprecated bioframe.split
    df1 = pd.DataFrame(
        [["chrX", 3, 8], ["chr1", 4, 7], ["chrX", 1, 5]],
        columns=["chrom", "start", "end"],
    )

    df2 = pd.DataFrame(
        [
            ["chrX", 4],
            ["chr1", 5],
        ],
        columns=["chrom", "pos"],
    )
    df2["start"] = df2["pos"]
    df2["end"] = df2["pos"]

    df_result = (
        pd.DataFrame(
            [
                ["chrX", 1, 4],
                ["chrX", 3, 4],
                ["chrX", 4, 5],
                ["chrX", 4, 8],
                ["chr1", 5, 7],
                ["chr1", 4, 5],
            ],
            columns=["chrom", "start", "end"],
        )
        .sort_values(["chrom", "start", "end"])
        .reset_index(drop=True)
        .astype({"start": pd.Int64Dtype(), "end": pd.Int64Dtype()})
    )

    pd.testing.assert_frame_equal(
        df_result,
        bioframe.subtract(df1, df2)
        .sort_values(["chrom", "start", "end"])
        .reset_index(drop=True),
    )

    # Test the case when a chromosome should not be split (now implemented with subtract)
    df1 = pd.DataFrame(
        [
            ["chrX", 3, 8],
            ["chr1", 4, 7],
        ],
        columns=["chrom", "start", "end"],
    )

    df2 = pd.DataFrame([["chrX", 4]], columns=["chrom", "pos"])
    df2["start"] = df2["pos"].values
    df2["end"] = df2["pos"].values

    df_result = (
        pd.DataFrame(
            [
                ["chrX", 3, 4],
                ["chrX", 4, 8],
                ["chr1", 4, 7],
            ],
            columns=["chrom", "start", "end"],
        )
        .sort_values(["chrom", "start", "end"])
        .reset_index(drop=True)
    )

    pd.testing.assert_frame_equal(
        df_result.astype({"start": pd.Int64Dtype(), "end": pd.Int64Dtype()}),
        bioframe.subtract(df1, df2)
        .sort_values(["chrom", "start", "end"])
        .reset_index(drop=True),
    )

    # subtract should ignore null rows
    df1 = pd.DataFrame(
        [[pd.NA, pd.NA, pd.NA], ["chr1", 1, 5]],
        columns=["chrom", "start", "end"],
    ).astype({"start": pd.Int64Dtype(), "end": pd.Int64Dtype()})

    df2 = pd.DataFrame(
        [
            ["chrX", 1, 5],
            [pd.NA, pd.NA, pd.NA],
            ["chr1", 4, 8],
            [pd.NA, pd.NA, pd.NA],
            ["chr1", 10, 11],
        ],
        columns=["chrom", "start", "end"],
    ).astype({"start": pd.Int64Dtype(), "end": pd.Int64Dtype()})

    df_subtracted = pd.DataFrame(
        [
            ["chr1", 1, 4],
        ],
        columns=["chrom", "start", "end"],
    ).astype({"start": pd.Int64Dtype(), "end": pd.Int64Dtype()})

    pd.testing.assert_frame_equal(df_subtracted, bioframe.subtract(df1, df2))

    df1 = pd.DataFrame(
        [
            [pd.NA, pd.NA, pd.NA],
        ],
        columns=["chrom", "start", "end"],
    ).astype({"start": pd.Int64Dtype(), "end": pd.Int64Dtype()})

    assert len(bioframe.subtract(df1, df2)) == 0  # empty df1 but valid chroms in df2

    with pytest.raises(ValueError):  # no non-null chromosomes
        bioframe.subtract(df1, df1)

    df2 = pd.DataFrame(
        [
            [pd.NA, pd.NA, pd.NA],
            [pd.NA, pd.NA, pd.NA],
        ],
        columns=["chrom", "start", "end"],
    ).astype({"start": pd.Int64Dtype(), "end": pd.Int64Dtype()})
    with pytest.raises(ValueError):  # no non-null chromosomes
        bioframe.subtract(df1, df2)


def test_setdiff():

    cols1 = ["chrom1", "start", "end"]
    cols2 = ["chrom2", "start", "end"]
    df1 = pd.DataFrame(
        [
            ["chr1", 8, 12, "+", "cat"],
            ["chr1", 8, 12, "-", "cat"],
            ["chrX", 1, 8, "+", "cat"],
        ],
        columns=cols1 + ["strand", "animal"],
    )
    df2 = pd.DataFrame(
        [
            ["chrX", 7, 10, "-", "dog"],
            ["chr1", 6, 10, "-", "cat"],
            ["chr1", 6, 10, "-", "cat"],
        ],
        columns=cols2 + ["strand", "animal"],
    )

    assert (
        len(
            bioframe.setdiff(
                df1,
                df2,
                cols1=cols1,
                cols2=cols2,
                on=None,
            )
        )
        == 0
    )  # everything overlaps

    assert (
        len(
            bioframe.setdiff(
                df1,
                df2,
                cols1=cols1,
                cols2=cols2,
                on=["animal"],
            )
        )
        == 1
    )  # two overlap, one remains

    assert (
        len(
            bioframe.setdiff(
                df1,
                df2,
                cols1=cols1,
                cols2=cols2,
                on=["strand"],
            )
        )
        == 2
    )  # one overlaps, two remain

    # setdiff should ignore nan rows
    df1 = pd.concat([pd.DataFrame([pd.NA]), df1, pd.DataFrame([pd.NA])])[
        ["chrom1", "start", "end", "strand", "animal"]
    ]
    df1 = df1.astype(
        {
            "start": pd.Int64Dtype(),
            "end": pd.Int64Dtype(),
        }
    )
    df2 = pd.concat([pd.DataFrame([pd.NA]), df2, pd.DataFrame([pd.NA])])[
        ["chrom2", "start", "end", "strand", "animal"]
    ]
    df2 = df2.astype(
        {
            "start": pd.Int64Dtype(),
            "end": pd.Int64Dtype(),
        }
    )

    assert (2, 5) == np.shape(bioframe.setdiff(df1, df1, cols1=cols1, cols2=cols1))
    assert (2, 5) == np.shape(bioframe.setdiff(df1, df2, cols1=cols1, cols2=cols2))
    assert (4, 5) == np.shape(
        bioframe.setdiff(df1, df2, on=["strand"], cols1=cols1, cols2=cols2)
    )


def test_count_overlaps():
    df1 = pd.DataFrame(
        [
            ["chr1", 8, 12, "+", "cat"],
            ["chr1", 8, 12, "-", "cat"],
            ["chrX", 1, 8, "+", "cat"],
        ],
        columns=["chrom1", "start", "end", "strand", "animal"],
    )

    df2 = pd.DataFrame(
        [
            ["chr1", 6, 10, "+", "dog"],
            ["chr1", 6, 10, "+", "dog"],
            ["chrX", 7, 10, "+", "dog"],
            ["chrX", 7, 10, "+", "dog"],
        ],
        columns=["chrom2", "start2", "end2", "strand", "animal"],
    )

    assert (
        bioframe.count_overlaps(
            df1,
            df2,
            on=None,
            cols1=("chrom1", "start", "end"),
            cols2=("chrom2", "start2", "end2"),
        )["count"].values
        == np.array([2, 2, 2])
    ).all()

    assert (
        bioframe.count_overlaps(
            df1,
            df2,
            on=["strand"],
            cols1=("chrom1", "start", "end"),
            cols2=("chrom2", "start2", "end2"),
        )["count"].values
        == np.array([2, 0, 2])
    ).all()

    assert (
        bioframe.count_overlaps(
            df1,
            df2,
            on=["strand", "animal"],
            cols1=("chrom1", "start", "end"),
            cols2=("chrom2", "start2", "end2"),
        )["count"].values
        == np.array([0, 0, 0])
    ).all()

    # overlaps with pd.NA
    counts_no_nans = bioframe.count_overlaps(
        df1,
        df2,
        on=None,
        cols1=("chrom1", "start", "end"),
        cols2=("chrom2", "start2", "end2"),
    )

    df1_na = (pd.concat([pd.DataFrame([pd.NA]), df1, pd.DataFrame([pd.NA])])).astype(
        {
            "start": pd.Int64Dtype(),
            "end": pd.Int64Dtype(),
        }
    )[["chrom1", "start", "end", "strand", "animal"]]
    df2_na = (pd.concat([pd.DataFrame([pd.NA]), df2, pd.DataFrame([pd.NA])])).astype(
        {
            "start2": pd.Int64Dtype(),
            "end2": pd.Int64Dtype(),
        }
    )[["chrom2", "start2", "end2", "strand", "animal"]]

    counts_nans_inserted_after = (
        pd.concat([pd.DataFrame([pd.NA]), counts_no_nans, pd.DataFrame([pd.NA])])
    ).astype({"start": pd.Int64Dtype(), "end": pd.Int64Dtype(),})[
        ["chrom1", "start", "end", "strand", "animal", "count"]
    ]

    counts_nans = bioframe.count_overlaps(
        df1_na,
        df2_na,
        on=None,
        cols1=("chrom1", "start", "end"),
        cols2=("chrom2", "start2", "end2"),
    )

    pd.testing.assert_frame_equal(
        counts_nans,
        bioframe.count_overlaps(
            df1_na,
            df2,
            on=None,
            cols1=("chrom1", "start", "end"),
            cols2=("chrom2", "start2", "end2"),
        ),
    )

    assert (
        counts_nans["count"].values
        == counts_nans_inserted_after["count"].fillna(0).values
    ).all()

    ### coverage without return_input returns a single column dataFrame
    pd.testing.assert_frame_equal(
        bioframe.count_overlaps(
            df1_na,
            df2_na,
            cols1=("chrom1", "start", "end"),
            cols2=("chrom2", "start2", "end2"),
            return_input=False,
        ),
        pd.DataFrame(counts_nans["count"]),
    )


def test_assign_view():

    ## default assignment case
    view_df = pd.DataFrame(
        [
            ["chr11", 1, 8, "chr11p"],
        ],
        columns=["chrom", "start", "end", "name"],
    )
    df = pd.DataFrame(
        [
            ["chr11", 0, 10, "+"],
        ],
        columns=["chrom", "start", "end", "strand"],
    )
    df_assigned = pd.DataFrame(
        [
            ["chr11", 0, 10, "+", "chr11p"],
        ],
        columns=["chrom", "start", "end", "strand", "view_region"],
    )
    df_assigned = df_assigned.astype(
        {"chrom": str, "start": pd.Int64Dtype(), "end": pd.Int64Dtype()}
    )
    pd.testing.assert_frame_equal(df_assigned, bioframe.assign_view(df, view_df))

    # non-default columns in view
    funky_view = view_df.rename(columns={"chrom": "chr"})
    pd.testing.assert_frame_equal(
        df_assigned,
        bioframe.assign_view(
            df,
            view_df=funky_view,
            cols_view=["chr", "start", "end"],
        ),
    )

    # assignment with funny view_name_col and an interval on chr2 not cataloged in the view_df
    view_df = pd.DataFrame(
        [
            ["chrX", 1, 8, "oranges"],
            ["chrX", 8, 20, "grapefruit"],
            ["chr1", 0, 10, "apples"],
        ],
        columns=["chrom", "start", "end", "fruit"],
    )

    df = pd.DataFrame(
        [
            ["chr1", 0, 10, "+"],
            ["chrX", 5, 10, "+"],
            ["chrX", 0, 5, "+"],
            ["chr2", 5, 10, "+"],
        ],
        columns=["chrom", "start", "end", "strand"],
    )

    df_assigned = pd.DataFrame(
        [
            ["chr1", 0, 10, "+", "apples"],
            ["chrX", 5, 10, "+", "oranges"],
            ["chrX", 0, 5, "+", "oranges"],
        ],
        columns=["chrom", "start", "end", "strand", "funny_view_region"],
    )
    df_assigned = df_assigned.astype(
        {"chrom": str, "start": pd.Int64Dtype(), "end": pd.Int64Dtype()}
    )

    pd.testing.assert_frame_equal(
        df_assigned,
        bioframe.assign_view(
            df,
            view_df,
            view_name_col="fruit",
            df_view_col="funny_view_region",
            drop_unassigned=True,
        ),
    )

    ### keep the interval with NA as its region if drop_unassigned is False
    df_assigned = pd.DataFrame(
        [
            ["chr1", 0, 10, "+", "apples"],
            ["chrX", 5, 10, "+", "oranges"],
            ["chrX", 0, 5, "+", "oranges"],
            ["chr2", 5, 10, "+", pd.NA],
        ],
        columns=["chrom", "start", "end", "strand", "funny_view_region"],
    )
    df_assigned = df_assigned.astype(
        {"chrom": str, "start": pd.Int64Dtype(), "end": pd.Int64Dtype()}
    )

    pd.testing.assert_frame_equal(
        df_assigned,
        bioframe.assign_view(
            df,
            view_df,
            view_name_col="fruit",
            df_view_col="funny_view_region",
            drop_unassigned=False,
        ),
    )

    ### assign_view with NA values assigns a view of none
    df = pd.DataFrame(
        [
            ["chr1", 0, 10, "+"],
            ["chrX", 5, 10, "+"],
            [pd.NA, pd.NA, pd.NA, "+"],
            ["chrX", 0, 5, "+"],
            ["chr2", 5, 10, "+"],
        ],
        columns=["chrom", "start", "end", "strand"],
    ).astype({"start": pd.Int64Dtype(), "end": pd.Int64Dtype()})

    pd.testing.assert_frame_equal(
        df, bioframe.assign_view(df, view_df, view_name_col="fruit").iloc[:, :-1]
    )

    assert (
        bioframe.assign_view(df, view_df, view_name_col="fruit")["view_region"].values
        == np.array(["apples", "oranges", None, "oranges", None], dtype=object)
    ).all()


def test_sort_bedframe():

    view_df = pd.DataFrame(
        [
            ["chrX", 1, 8, "oranges"],
            ["chrX", 8, 20, "grapefruit"],
            ["chr1", 0, 10, "apples"],
        ],
        columns=["chrom", "start", "end", "fruit"],
    )

    df = pd.DataFrame(
        [
            ["chr2", 5, 10, "+"],
            ["chr1", 0, 10, "+"],
            ["chrX", 5, 10, "+"],
            ["chrX", 0, 5, "+"],
        ],
        columns=["chrom", "start", "end", "strand"],
    )

    # sorting just by chrom,start,end
    df_sorted = pd.DataFrame(
        [
            ["chr1", 0, 10, "+"],
            ["chr2", 5, 10, "+"],
            ["chrX", 0, 5, "+"],
            ["chrX", 5, 10, "+"],
        ],
        columns=["chrom", "start", "end", "strand"],
    )

    pd.testing.assert_frame_equal(df_sorted, bioframe.sort_bedframe(df))

    # when a view_df is provided, regions without assigned views
    # are placed last
    df_sorted = pd.DataFrame(
        [
            ["chrX", 0, 5, "+"],
            ["chrX", 5, 10, "+"],
            ["chr1", 0, 10, "+"],
            ["chr2", 5, 10, "+"],
        ],
        columns=["chrom", "start", "end", "strand"],
    )

    # test if sorting after assiging view to df denovo works,
    pd.testing.assert_frame_equal(
        df_sorted, bioframe.sort_bedframe(df, view_df, view_name_col="fruit")
    )

    # non-default columns in view
    funky_view = view_df.rename(columns={"chrom": "chr"})
    pd.testing.assert_frame_equal(
        df_sorted,
        bioframe.sort_bedframe(
            df,
            view_df=funky_view,
            view_name_col="fruit",
            cols_view=["chr", "start", "end"],
        ),
    )

    # also test if sorting after assigning view to df denovo works with default view_name_col
    pd.testing.assert_frame_equal(
        df_sorted, bioframe.sort_bedframe(df, view_df.rename(columns={"fruit": "name"}))
    )

    # also test if sorting after assiging view to df from chromsizes-like dictionary works:
    pd.testing.assert_frame_equal(
        df_sorted, bioframe.sort_bedframe(df, view_df={"chrX": 20, "chr1": 10})
    )

    ### 'df' has no column 'view_region', so this should raise a ValueError
    assert pytest.raises(
        ValueError,
        bioframe.sort_bedframe,
        df,
        view_df,
        view_name_col="fruit",
        df_view_col="view_region",
    )

    ### sort_bedframe with NA entries:
    df = pd.DataFrame(
        [
            ["chr1", 0, 10, "+"],
            ["chrX", 5, 10, "+"],
            [pd.NA, pd.NA, pd.NA, "+"],
            ["chrX", 0, 5, "+"],
            ["chr2", 5, 10, "+"],
        ],
        columns=["chrom", "start", "end", "strand"],
    ).astype({"start": pd.Int64Dtype(), "end": pd.Int64Dtype()})

    # NA put at end
    assert pd.isna(bioframe.sort_bedframe(df)["chrom"].values[-1])
    assert pd.isna(
        bioframe.sort_bedframe(df, view_df, view_name_col="fruit")["chrom"].values[-1]
    )
    assert (
        df.dtypes == bioframe.sort_bedframe(df, view_df, view_name_col="fruit").dtypes
    ).all()



import pandas as pd
import numpy as np
from . import construction
from .specs import _get_default_colnames, _verify_columns, _verify_column_dtypes
from .. import ops

__all__ = [
    "is_bedframe",
    "is_cataloged",
    "is_overlapping",
    "is_viewframe",
    "is_contained",
    "is_covering",
    "is_tiling",
    "is_sorted",
]


def is_bedframe(
    df,
    raise_errors=False,
    cols=None,
):
    """
    Checks that required bedframe properties are satisfied for dataframe `df`.

    This includes:

    - chrom, start, end columns
    - columns have valid dtypes (object/string/categorical, int/pd.Int64Dtype, int/pd.Int64Dtype)
    - for each interval, if any of chrom, start, end are null, then all are null
    - all starts < ends.

    Parameters
    ----------
    df : pandas.DataFrame

    raise_errors : bool
        If True, raises errors instead of returning a boolean False for invalid properties.
        Default False.

    cols : (str, str, str) or None
        The names of columns containing the chromosome, start and end of the
        genomic intervals, provided separately for each set. The default
        values are 'chrom', 'start', 'end'.

    Returns
    -------
    is_bedframe:bool

    """
    ck1, sk1, ek1 = _get_default_colnames() if cols is None else cols

    if not _verify_columns(df, [ck1, sk1, ek1], return_as_bool=True):
        if raise_errors:
            raise TypeError("Invalid bedFrame: Invalid column names")
        return False

    if not _verify_column_dtypes(df, cols=[ck1, sk1, ek1], return_as_bool=True):
        if raise_errors:
            raise TypeError("Invalid bedFrame: Invalid column dtypes")
        return False

    nan_intervals = pd.isnull(df[[ck1, sk1, ek1]])
    if (~(~nan_intervals.any(axis=1) | nan_intervals.all(axis=1))).any():
        if raise_errors:
            raise ValueError(
                "Invalid bedFrame: Invalid null values (if any of chrom, start, end are null, then each must be null)"
            )
        return False

    if ((df[ek1] - df[sk1]) < 0).any():
        if raise_errors:
            raise ValueError(
                "Invalid bedFrame: starts exceed ends for "
                + str(np.sum(((df[ek1] - df[sk1]) < 0)))
                + " intervals"
            )
        return False

    return True


def is_cataloged(
    df, view_df, raise_errors=False, df_view_col="view_region", view_name_col="name"
):
    """
    Tests if all region names in `df[df_view_col]` are present in `view_df[view_name_col]`.

    Parameters
    ----------
    df : pandas.DataFrame

    view_df : pandas.DataFrame

    raise_errors : bool
        If True, raises errors instead of returning a boolean False for invalid properties.
        Default False.

    df_view_col: str
        Name of column from df that indicates region in view.

    view_name_col: str
        Name of column from view that specifies  region name.

    Returns
    -------
    is_cataloged:bool

    Notes
    -----
    Does not check if names in `view_df[view_name_col]` are unique.

    """
    if not _verify_columns(df, [df_view_col], return_as_bool=True):
        if raise_errors is True:
            raise ValueError(f"Could not find ‘{df_view_col}’ column in df")
        return False

    if not _verify_columns(view_df, [view_name_col], return_as_bool=True):
        if raise_errors is True:
            raise ValueError(f"Could not find ‘{view_name_col}’ column in view_df")
        return False

    if not set(df[df_view_col].copy().dropna().values).issubset(
        set(view_df[view_name_col].values)
    ):
        if raise_errors is True:
            raise ValueError(
                "The following regions in df[df_view_col] not in view_df[view_name_col]: \n"
                + "{}".format(
                    set(df[df_view_col].values).difference(
                        set(view_df[view_name_col].values)
                    )
                )
            )
        return False

    return True


def is_overlapping(df, cols=None):
    """
    Tests if any genomic intervals in a bioframe `df` overlap.

    Also see :func:`bioframe.ops.merge()`.

    Parameters
    ----------
    df : pandas.DataFrame

    cols : (str, str, str) or None
        The names of columns containing the chromosome, start and end of the
        genomic intervals, provided separately for each set. The default
        values are 'chrom', 'start', 'end'.

    Returns
    -------
    is_overlapping:bool

    """
    from ..ops import merge

    ck1, sk1, ek1 = _get_default_colnames() if cols is None else cols

    df_merged = merge(df, cols=cols)

    total_interval_len = np.sum((df[ek1] - df[sk1]).values)
    total_interval_len_merged = np.sum((df_merged[ek1] - df_merged[sk1]).values)

    if total_interval_len > total_interval_len_merged:
        return True
    else:
        return False


def is_viewframe(region_df, raise_errors=False, view_name_col="name", cols=None):
    """
    Checks that `region_df` is a valid viewFrame.

    This includes:

    - it satisfies requirements for a bedframe, including columns for ('chrom', 'start', 'end')
    - it has an additional column, view_name_col, with default 'name'
    - it does not contain null values
    - entries in the view_name_col are unique.
    - intervals are non-overlapping

    Parameters
    ----------

    region_df : pandas.DataFrame
        Dataframe of genomic intervals to be tested.

    raise_errors : bool
        If True, raises errors instead of returning a boolean False for invalid properties.
        Default False.

    view_name_col : str
        Specifies column name of the view regions. Default 'name'.

    cols : (str, str, str) or None
        The names of columns containing the chromosome, start and end of the
        genomic intervals, provided separately for each set. The default
        values are 'chrom', 'start', 'end'.

    Returns
    -------
    is_viewframe:bool

    """

    ck1, sk1, ek1 = _get_default_colnames() if cols is None else cols

    if not _verify_columns(
        region_df, [ck1, sk1, ek1, view_name_col], return_as_bool=True
    ):
        if raise_errors:
            raise TypeError("Invalid view: invalid column names")
        return False

    if not is_bedframe(region_df, cols=cols):
        if raise_errors:
            raise ValueError("Invalid view: not a bedframe")
        return False

    if pd.isna(region_df).values.any():
        if raise_errors:
            raise ValueError("Invalid view: cannot contain NAs")
        return False

    if len(set(region_df[view_name_col])) < len(region_df[view_name_col].values):
        if raise_errors:
            raise ValueError(
                "Invalid view: entries in region_df[view_name_col] must be unique"
            )
        return False

    if is_overlapping(region_df, cols=cols):
        if raise_errors:
            raise ValueError("Invalid view: entries must be non-overlapping")
        return False

    return True


def is_contained(
    df,
    view_df,
    raise_errors=False,
    df_view_col=None,
    view_name_col="name",
    cols=None,
):
    """
    Tests if all genomic intervals in a bioframe `df` are cataloged and do not extend beyond their
    associated region in the view `view_df`.

    Parameters
    ----------
    df : pandas.DataFrame

    view_df : pandas.DataFrame
        Valid viewframe.

    raise_errors : bool
        If True, raises errors instead of returning a boolean False for invalid properties.
        Default False.

    df_view_col:
        Column from df used to associate interviews with view regions.
        Default `view_region`.

    cols: (str, str, str)
        Column names for chrom, start, end in df.

    Returns
    -------
    is_contained:bool

    """
    from ..ops import trim

    ck1, sk1, ek1 = _get_default_colnames() if cols is None else cols

    if df_view_col is None:
        try:
            df_view_assigned = ops.overlap(df, view_df)
            assert (df_view_assigned["end_"].isna()).sum() == 0
            assert (df_view_assigned["start_"].isna()).sum() == 0
            assert (df_view_assigned["end"] <= df_view_assigned["end_"]).all()
            assert (df_view_assigned["start"] >= df_view_assigned["start_"]).all()
        except AssertionError:
            if raise_errors:
                raise AssertionError("df not contained in view_df")
            else:
                return False
        return True

    if not is_cataloged(
        df, view_df, df_view_col=df_view_col, view_name_col=view_name_col
    ):
        if raise_errors:
            raise ValueError("df not cataloged in view_df")
        return False

    df_trim = trim(
        df, view_df=view_df, df_view_col=df_view_col, view_name_col=view_name_col
    )
    is_start_trimmed = np.any(df[sk1].values != df_trim[sk1].values)
    is_end_trimmed = np.any(df[ek1].values != df_trim[ek1].values)

    if is_start_trimmed or is_end_trimmed:
        if raise_errors:
            raise ValueError("df not contained in view_df")
        return False
    else:
        return True


def is_covering(df, view_df, view_name_col="name", cols=None):
    """
    Tests if a view `view_df` is covered by the set of genomic intervals in the bedframe `df`.

    This test is true if ``complement(df,view_df)`` is empty. Also note this test ignores regions assigned to
    intervals in `df` since regions are re-assigned in :func:`bioframe.ops.complement`.

    Parameters
    ----------
    df : pandas.DataFrame

    view_df : pandas.DataFrame
        Valid viewFrame.

    view_name_col:
        Column from view_df with view region names. Default `name`.

    cols : (str, str, str) or None
        The names of columns containing the chromosome, start and end of the
        genomic intervals, provided separately for each set. The default
        values are 'chrom', 'start', 'end'.

    Returns
    -------
    is_covering:bool

    """
    from ..ops import complement

    if complement(
        df,
        view_df=view_df,
        view_name_col=view_name_col,
        cols=cols,
    ).empty:
        return True
    else:
        return False


def is_tiling(
    df,
    view_df,
    raise_errors=False,
    df_view_col="view_region",
    view_name_col="name",
    cols=None,
):
    """
    Tests if a view `view_df` is tiled by the set of genomic intervals in the bedframe `df`.

    This is true if:

    - df is not overlapping
    - df is covering view_df
    - df is contained in view_df

    Parameters
    ----------
    df : pandas.DataFrame

    view_df : pandas.DataFrame
        valid viewFrame

    raise_errors : bool
        If True, raises errors instead of returning a boolean False for invalid properties.
        Default False.

    df_view_col: str
        Name of column from df that indicates region in view.

    view_name_col: str
        Name of column from view that specifies unique region name.

    cols : (str, str, str) or None
        The names of columns containing the chromosome, start and end of the
        genomic intervals, provided separately for each set. The default
        values are 'chrom', 'start', 'end'.

    Returns
    -------
    is_tiling:bool

    """

    view_df = construction.make_viewframe(
        view_df, view_name_col=view_name_col, cols=cols
    )

    if is_overlapping(df):
        if raise_errors:
            raise ValueError("overlaps")
        return False
    if not is_covering(df, view_df, view_name_col=view_name_col, cols=None):
        if raise_errors:
            raise ValueError("not covered")
        return False
    if not is_contained(
        df, view_df, df_view_col=df_view_col, view_name_col=view_name_col, cols=None
    ):
        if raise_errors:
            raise ValueError("not contained")
        return False
    return True


def is_sorted(
    df,
    view_df=None,
    reset_index=True,
    df_view_col=None,
    view_name_col="name",
    cols=None,
):
    """
    Tests if a bedframe is changed by sorting.

    Also see :func:`bioframe.ops.sort_bedframe`.

    Parameters
    ----------
    df : pandas.DataFrame

    view_df : pandas.DataFrame | dict-like
        Optional view to pass to ``sort_bedframe``.
        When it is dict-like :func:'bioframe.make_viewframe' will
        be used to convert to viewframe. If view_df is not provided
        df is assumed to be sorted by chrom and start.

    reset_index : bool
        Optional argument to pass to ``sort_bedframe``.

    df_view_col: None | str
        Name of column from df that indicates region in view.
        If None, :func:'bioframe.assign_view' will be used to assign view regions.
        Default None.

    view_name_col: str
        Name of column from view that specifies unique region name.

    cols : (str, str, str) or None
        The names of columns containing the chromosome, start and end of the
        genomic intervals, provided separately for each set. The default
        values are 'chrom', 'start', 'end'.

    Returns
    -------
    is_sorted : bool

    """
    from ..ops import sort_bedframe

    df_sorted = sort_bedframe(
        df.copy(),
        view_df=view_df,
        reset_index=reset_index,
        df_view_col=df_view_col,
        view_name_col=view_name_col,
        cols=cols,
    )

    if df.equals(df_sorted):
        return True
    else:
        return False



from .schemas import SCHEMAS

from . import fileops
from .fileops import *

from . import resources
from .resources import *

__all__ = [
    "SCHEMAS",
    *fileops.__all__,
    *resources.__all__,
]



from functools import partial
from urllib.parse import urljoin, urlencode
import urllib
import os
import posixpath as pp
import os.path as op
import pandas as pd
import requests
import socket
import base64
import glob

import pkg_resources

from .schemas import SCHEMAS, UCSC_MRNA_FIELDS
from .fileops import (
    read_table,
    read_chromsizes,
    extract_centromeres,
)

__all__ = [
    "fetch_chromsizes",
    "fetch_centromeres",
    "UCSCClient",
]


def _check_connectivity(reference="http://www.google.com"):
    try:
        urllib.request.urlopen(reference, timeout=5)
        return True
    except urllib.request.URLError:
        return False
    except socket.timeout:
        return False


def fetch_chromsizes(
    db,
    provider=None,
    filter_chroms=True,
    chrom_patterns=(r"^chr[0-9]+$", r"^chr[XY]$", r"^chrM$"),
    natsort=True,
    as_bed=False,
    **kwargs,
):
    """
    Fetch chromsizes from the UCSC database or local storage.

    Parameters
    ----------
    provider : str
        The provider of chromsizes. Currently, only 'ucsc' is implemented.
    filter_chroms : bool, optional
        Filter for chromosome names given in ``chrom_patterns``.
    chrom_patterns : sequence, optional
        Sequence of regular expressions to capture desired sequence names.
    natsort : bool, optional
        Sort each captured group of names in natural order. Default is True.
    as_bed : bool, optional
        If True, return chromsizes as an interval dataframe (chrom, start, end).
    **kwargs :
        Passed to :func:`pandas.read_csv`

    Returns
    -------
    Series of integer bp lengths indexed by sequence name or an interval dataframe.

    """

    if provider == "local":
        fpath = f"data/{db}.chrom.sizes"
        if pkg_resources.resource_exists("bioframe.io", fpath):
            return read_chromsizes(
                pkg_resources.resource_filename("bioframe.io", fpath)
            )
        else:
            raise LookupError(f"Assembly '{db}' not found in local storage")

    if provider == "ucsc" or provider is None:
        return UCSCClient(db).fetch_chromsizes(
            filter_chroms=filter_chroms,
            chrom_patterns=chrom_patterns,
            natsort=natsort,
            as_bed=as_bed,
            **kwargs,
        )
    else:
        raise ValueError("Unknown provider '{}'".format(provider))


def fetch_centromeres(db, provider=None, merge=True, verbose=False):
    """
    Extract centromere locations for a given assembly 'db' from a variety
    of file formats in UCSC (centromeres, cytoband, gap) depending on
    availability, returning a DataFrame.

    Parameters
    ----------

    db : str

    merge : bool
        Whether to merge all centromere intervals per chromosome into
        one consolidated centromere interval.
        Default True.

    Returns
    -------
    DataFrame with centromere 'chrom', 'start', 'end', 'mid'.

    Notes
    -----
    The priority goes as
    - Local (not implemented)
    - centromeres.txt
    - cytoBandIdeo
    - cytoBand
    - gap.txt

    Currently only works for human assemblies.


    """

    if provider == "local":
        raise NotImplementedError("local method not currently implemented")
        fpath = f"data/{db}.centromeres"
        if pkg_resources.resource_exists("bioframe.io", fpath):
            return read_chromsizes(
                pkg_resources.resource_filename("bioframe.io", fpath)
            )
        else:
            raise LookupError(f"Centromeres for '{db}' not found in local storage")

    if provider == "ucsc" or provider is None:
        client = UCSCClient(db)
        fetchers = [
            ("centromeres", client.fetch_centromeres),
            ("cytoband", client.fetch_cytoband),
            ("cytoband", partial(client.fetch_cytoband, ideo=True)),
            ("gap", client.fetch_gaps),
        ]

        for schema, fetcher in fetchers:
            try:
                df = fetcher()
                break
            except urllib.error.HTTPError:
                pass
        else:
            raise ValueError("No source for centromere data found.")
    else:
        raise NotImplementedError("currently UCSC is only implemented provider")

    return extract_centromeres(df, schema=schema, merge=merge)


class UCSCClient:
    BASE_URL = "http://hgdownload.cse.ucsc.edu/"

    def __init__(self, db):
        self._db = db
        self._db_url = urljoin(self.BASE_URL, "goldenPath/{}/database/".format(db))

    def fetch_chromsizes(
        self,
        filter_chroms=True,
        chrom_patterns=(r"^chr[0-9]+$", r"^chr[XY]$", r"^chrM$"),
        natsort=True,
        as_bed=False,
        **kwargs,
    ):
        url = urljoin(self._db_url, "chromInfo.txt.gz")
        return read_chromsizes(
            url,
            filter_chroms=filter_chroms,
            chrom_patterns=chrom_patterns,
            natsort=natsort,
            as_bed=as_bed,
            **kwargs,
        )

    def fetch_centromeres(self, **kwargs):
        url = urljoin(self._db_url, "centromeres.txt.gz")
        return read_table(url, schema="centromeres")

    def fetch_gaps(self, **kwargs):
        url = urljoin(self._db_url, "gap.txt.gz")
        return read_table(
            url,
            schema="gap",
            usecols=["chrom", "start", "end", "length", "type", "bridge"],
            **kwargs,
        )

    def fetch_cytoband(self, ideo=False, **kwargs):
        if ideo:
            url = urljoin(self._db_url, "cytoBandIdeo.txt.gz")
        else:
            url = urljoin(self._db_url, "cytoBand.txt.gz")
        return read_table(url, schema="cytoband")
    
    def fetch_mrna(self, **kwargs):
        url = urljoin(self._db_url, "all_mrna.txt.gz")
        return read_table(
            url,
            schema=UCSC_MRNA_FIELDS,
            **kwargs,
        )



from collections import OrderedDict
from contextlib import closing
import subprocess
import tempfile
import json
import io

import numpy as np
import pandas as pd

try:
    import bbi
except ImportError:
    bbi = None

try:
    import pyBigWig
except ImportError:
    pyBigWig = None

from ..core.stringops import parse_region
from ..core.arrops import argnatsort
from .schemas import SCHEMAS, BAM_FIELDS


__all__ = [
    "read_table",
    "read_chromsizes",
    "read_tabix",
    "read_pairix",
    "read_bam",
    "load_fasta",
    "read_bigwig",
    "to_bigwig",
    "read_bigbed",
    "to_bigbed",
]


def read_table(filepath_or, schema=None, schema_is_strict=False, **kwargs):
    """
    Read a tab-delimited file into a data frame.

    Equivalent to :func:`pandas.read_table` but supports an additional
    `schema` argument to populate column names for common genomic formats.

    Parameters
    ----------
    filepath_or : str, path object or file-like object
        Any valid string path is acceptable. The string could be a URL
    schema : str
        Schema to use for table column names.
    schema_is_strict : bool
        Whether to check if columns are filled with NAs.

    Returns
    -------
    df : pandas.DataFrame of intervals

    """
    kwargs.setdefault("sep", "\t")
    kwargs.setdefault("header", None)
    kwargs.setdefault("index_col", False)
    if isinstance(filepath_or, str) and filepath_or.endswith(".gz"):
        kwargs.setdefault("compression", "gzip")
    if schema is not None:
        try:
            kwargs.setdefault("names", SCHEMAS[schema])
        except (KeyError, TypeError):
            if isinstance(schema, str):
                raise ValueError("TSV schema not found: '{}'".format(schema))
            kwargs.setdefault("names", schema)
    df = pd.read_csv(filepath_or, **kwargs)
    if schema_is_strict:
        if (df.notna().sum(axis=0) == 0).any():
            raise ValueError(
                "one or more columns are all NA,"
                + " check agreement between number of fields in schema"
                + " and number of columns in input file"
            )
    return df


def read_chromsizes(
    filepath_or,
    filter_chroms=True,
    chrom_patterns=(r"^chr[0-9]+$", r"^chr[XY]$", r"^chrM$"),
    natsort=True,
    as_bed=False,
    **kwargs
):
    """
    Read a ``<db>.chrom.sizes`` or ``<db>.chromInfo.txt`` file from the UCSC
    database, where ``db`` is a genome assembly name, as a `pandas.Series`.

    Parameters
    ----------
    filepath_or : str or file-like
        Path or url to text file, or buffer.
    filter_chroms : bool, optional
        Filter for chromosome names given in ``chrom_patterns``.
    chrom_patterns : sequence, optional
        Sequence of regular expressions to capture desired sequence names.
    natsort : bool, optional
        Sort each captured group of names in natural order. Default is True.
    as_bed : bool, optional
        If True, return chromsizes as an interval dataframe (chrom, start, end).
    **kwargs :
        Passed to :func:`pandas.read_csv`

    Returns
    -------
    Series of integer bp lengths indexed by sequence name or an interval dataframe.

    Notes
    -----
    Mention name patterns

    See also
    --------
    * UCSC assembly terminology: <http://genome.ucsc.edu/FAQ/FAQdownloads.html#download9>
    * NCBI assembly terminology: <https://www.ncbi.nlm.nih.gov/grc/help/definitions>

    """
    if isinstance(filepath_or, str) and filepath_or.endswith(".gz"):
        kwargs.setdefault("compression", "gzip")

    chromtable = pd.read_csv(
        filepath_or,
        sep="\t",
        usecols=[0, 1],
        names=["name", "length"],
        dtype={"name": str},
        **kwargs
    )

    if filter_chroms:
        parts = []
        for pattern in chrom_patterns:
            if not len(pattern):
                continue
            part = chromtable[chromtable["name"].str.contains(pattern)]
            if natsort:
                part = part.iloc[argnatsort(part["name"])]
            parts.append(part)
        chromtable = pd.concat(parts, axis=0)

    if as_bed:
        chromtable["start"] = 0
        chromtable = (
            chromtable[["name", "start", "length"]]
            .rename({"name": "chrom", "length": "end"}, axis="columns")
            .reset_index(drop=True)
        )
    else:
        chromtable.index = chromtable["name"].values
        chromtable = chromtable["length"]
    return chromtable


def read_tabix(fp, chrom=None, start=None, end=None):
    """
    Read a tabix-indexed file into dataFrame.
    """
    import pysam

    with closing(pysam.TabixFile(fp)) as f:
        names = list(f.header) or None
        df = pd.read_csv(
            io.StringIO("\n".join(f.fetch(chrom, start, end))),
            sep="\t",
            header=None,
            names=names,
        )
    return df


def read_pairix(
    fp,
    region1,
    region2=None,
    chromsizes=None,
    columns=None,
    usecols=None,
    dtypes=None,
    **kwargs
):
    """
    Read a pairix-indexed file into DataFrame.
    """
    import pypairix
    import cytoolz as toolz

    if dtypes is None:
        dtypes = {}
    f = pypairix.open(fp, "r")

    header = f.get_header()
    if len(header):
        header_groups = toolz.groupby(lambda x: x.split(":")[0], header)
        if "#chromsize" in header_groups and chromsizes is None:
            items = [line.split()[1:] for line in header_groups["#chromsize"]]
            if len(items) and chromsizes is None:
                names, lengths = zip(*((item[0], int(item[1])) for item in items))
                chromsizes = pd.Series(index=names, data=lengths)
        if "#columns" in header_groups and columns is None:
            columns = header_groups["#columns"][0].split()[1:]

    chrom1, start1, end1 = parse_region(region1, chromsizes)
    if region2 is not None:
        chrom2, start2, end2 = parse_region(region2, chromsizes)
    else:
        chrom2, start2, end2 = chrom1, start1, end1

    it = f.query2D(chrom1, start1, end1, chrom2, start2, end2)
    if usecols is not None:
        argusecols = [columns.index(col) for col in usecols]
        records = [(record[i] for i in argusecols) for record in it]
        columns = usecols
    else:
        records = it

    df = pd.DataFrame.from_records(records, columns=columns)
    if columns is not None:
        for col in columns:
            if col in dtypes:
                df[col] = df[col].astype(dtypes[col])
            else:
                df[col] = pd.to_numeric(df[col], "ignore")
    return df


def read_bam(fp, chrom=None, start=None, end=None):
    """
    Read bam records into a DataFrame.
    """
    import pysam

    with closing(pysam.AlignmentFile(fp, "rb")) as f:
        bam_iter = f.fetch(chrom, start, end)
        records = [
            (
                s.qname,
                s.flag,
                s.rname,
                s.pos,
                s.mapq,
                s.cigarstring if s.mapq != 0 else np.nan,
                s.rnext,
                s.pnext,
                s.tlen,
                s.seq,
                s.qual,
                json.dumps(OrderedDict(s.tags)),
            )
            for s in bam_iter
        ]
        df = pd.DataFrame(records, columns=BAM_FIELDS)
    return df


def extract_centromeres(df, schema=None, merge=True):
    """
    Attempts to extract centromere locations from a variety of file formats,
    returning 'chrom', 'start', 'end', 'mid' in a pandas.DataFrame.
    """

    if schema == "centromeres":
        cens = df
    elif schema == "cytoband":
        cens = df[df["gieStain"] == "acen"]
    elif schema == "gap":
        cens = df[df["type"] == "centromere"]
    else:
        raise ValueError('`schema` must be one of {"centromeres", "cytoband", "gap"}.')

    if merge:
        cens = cens.groupby("chrom").agg({"start": np.min, "end": np.max}).reset_index()

    cens["mid"] = (cens["start"] + cens["end"]) // 2
    cens = (
        cens[["chrom", "start", "end", "mid"]]
        .sort_values("chrom")
        .reset_index(drop=True)
    )

    return cens


class PysamFastaRecord(object):
    def __init__(self, ff, ref):
        self.ff = ff
        if ref not in ff.references:
            raise KeyError("Reference name '{}' not found in '{}'".format(ref, ff))
        self.ref = ref

    def __getitem__(self, key):
        if isinstance(key, slice):
            start, stop = key.start, key.stop
        else:
            start = key
            stop = key + 1
        return self.ff.fetch(self.ref, start, stop)


def load_fasta(filepath_or, engine="pysam", **kwargs):
    """
    Load lazy fasta sequences from an indexed fasta file (optionally compressed)
    or from a collection of uncompressed fasta files.

    Parameters
    ----------
    filepath_or : str or iterable
        If a string, a filepath to a single `.fa` or `.fa.gz` file. Assumed to
        be accompanied by a `.fai` index file. Depending on the engine, the
        index may be created on the fly, and some compression formats may not
        be supported. If not a string, an iterable of fasta file paths each
        assumed to contain a single sequence.
    engine : {'pysam', 'pyfaidx'}, optional
        Module to use for loading sequences.
    kwargs : optional
        Options to pass to ``pysam.FastaFile`` or ``pyfaidx.Fasta``.

    Returns
    -------
    OrderedDict of (lazy) fasta records.

    Notes
    -----
    * pysam/samtools can read .fai and .gzi indexed files, I think.
    * pyfaidx can handle uncompressed and bgzf compressed files.

    """
    is_multifile = not isinstance(filepath_or, str)
    records = OrderedDict()

    engine = engine.lower()

    if engine == "pysam":
        try:
            import pysam
        except ImportError:
            raise ImportError("pysam is required to use engine='pysam'")

        if is_multifile:
            for onefile in filepath_or:
                ff = pysam.FastaFile(onefile, **kwargs)
                name = ff.references[0]
                records[name] = PysamFastaRecord(ff, name)
        else:
            ff = pysam.FastaFile(filepath_or, **kwargs)
            for name in ff.references:
                records[name] = PysamFastaRecord(ff, name)

    elif engine == "pyfaidx":
        try:
            import pyfaidx
        except ImportError:
            raise ImportError("pyfaidx is required to use engine='pyfaidx'")

        if is_multifile:
            for onefile in filepath_or:
                ff = pyfaidx.Fasta(onefile, **kwargs)
                name = next(iter(ff.keys()))
                records[name] = ff[name]
        else:
            ff = pyfaidx.Fasta(filepath_or, **kwargs)
            for name in ff.keys():
                records[name] = ff[name]

    else:
        raise ValueError("engine must be 'pysam' or 'pyfaidx'")

    return records


def read_bigwig(path, chrom, start=None, end=None, engine="auto"):
    """
    Read intervals from a bigWig file.

    Parameters
    ----------
    path : str
        Path or URL to a bigWig file
    chrom : str
    start, end : int, optional
        Start and end coordinates. Defaults to 0 and chromosome length.
    engine : {"auto", "pybbi", "pybigwig"}
        Library to use for querying the bigWig file.

    Returns
    -------
    DataFrame

    """
    engine = engine.lower()

    if engine == "auto":
        if bbi is None and pyBigWig is None:
            raise ImportError(
                "read_bigwig requires either the pybbi or pyBigWig package"
            )
        elif bbi is not None:
            engine = "pybbi"
        else:
            engine = "pybigwig"

    if engine in ("pybbi", "bbi"):
        if start is None:
            start = 0
        if end is None:
            end = -1
        with bbi.open(path) as f:
            df = f.fetch_intervals(chrom, start=start, end=end)

    elif engine == "pybigwig":
        f = pyBigWig.open(path)
        if start is None:
            start = 0
        if end is None:
            end = f.chroms()[chrom]
        ivals = f.intervals(chrom, start, end)
        df = pd.DataFrame(ivals, columns=["start", "end", "value"])
        df.insert(0, "chrom", chrom)

    else:
        raise ValueError(
            "engine must be 'auto', 'pybbi' or 'pybigwig'; got {}".format(engine)
        )

    return df


def read_bigbed(path, chrom, start=None, end=None, engine="auto"):
    """
    Read intervals from a bigBed file.

    Parameters
    ----------
    path : str
        Path or URL to a bigBed file
    chrom : str
    start, end : int, optional
        Start and end coordinates. Defaults to 0 and chromosome length.
    engine : {"auto", "pybbi", "pybigwig"}
        Library to use for querying the bigBed file.

    Returns
    -------
    DataFrame

    """
    engine = engine.lower()

    if engine == "auto":
        if bbi is None and pyBigWig is None:
            raise ImportError(
                "read_bigbed requires either the pybbi or pyBigWig package"
            )
        elif bbi is not None:
            engine = "pybbi"
        else:
            engine = "pybigwig"

    if engine in ("pybbi", "bbi"):
        if start is None:
            start = 0
        if end is None:
            end = -1
        with bbi.open(path) as f:
            df = f.fetch_intervals(chrom, start=start, end=end)

    elif engine == "pybigwig":
        f = pyBigWig.open(path)
        if start is None:
            start = 0
        if end is None:
            end = f.chroms()[chrom]
        ivals = f.entries(chrom, start, end)
        df = pd.DataFrame(ivals, columns=["start", "end", "rest"])
        df.insert(0, "chrom", chrom)

    else:
        raise ValueError(
            "engine must be 'auto', 'pybbi' or 'pybigwig'; got {}".format(engine)
        )

    return df


def to_bigwig(df, chromsizes, outpath, value_field=None):
    """
    Save a bedGraph-like dataframe as a binary BigWig track.

    Parameters
    ----------
    df : pandas.DataFrame
        Data frame with columns 'chrom', 'start', 'end' and one or more value
        columns
    chromsizes : pandas.Series
        Series indexed by chromosome name mapping to their lengths in bp
    outpath : str
        The output BigWig file path
    value_field : str, optional
        Select the column label of the data frame to generate the track. Default
        is to use the fourth column.

    """
    is_bedgraph = True
    for col in ["chrom", "start", "end"]:
        if col not in df.columns:
            is_bedgraph = False
    if len(df.columns) < 4:
        is_bedgraph = False

    if not is_bedgraph:
        raise ValueError(
            "A bedGraph-like DataFrame is required, got {}".format(df.columns)
        )

    if value_field is None:
        value_field = df.columns[3]

    columns = ["chrom", "start", "end", value_field]
    bg = df[columns].copy()
    bg["chrom"] = bg["chrom"].astype(str)
    bg = bg.sort_values(["chrom", "start", "end"])

    with tempfile.NamedTemporaryFile(suffix=".bg") as f, tempfile.NamedTemporaryFile(
        "wt", suffix=".chrom.sizes"
    ) as cs:

        chromsizes.to_csv(cs, sep="\t", header=False)
        cs.flush()

        bg.to_csv(
            f.name, sep="\t", columns=columns, index=False, header=False, na_rep="nan"
        )

        p = subprocess.run(
            ["bedGraphToBigWig", f.name, cs.name, outpath],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
    return p


def to_bigbed(df, chromsizes, outpath, schema="bed6"):
    """
    Save a bedGraph-like dataframe as a binary BigWig track.

    Parameters
    ----------
    df : pandas.DataFrame
        Data frame with columns 'chrom', 'start', 'end' and one or more value
        columns
    chromsizes : pandas.Series
        Series indexed by chromosome name mapping to their lengths in bp
    outpath : str
        The output BigWig file path
    value_field : str, optional
        Select the column label of the data frame to generate the track. Default
        is to use the fourth column.

    """
    is_bed6 = True
    for col in ["chrom", "start", "end", "name", "score", "strand"]:
        if col not in df.columns:
            is_bed6 = False
    if len(df.columns) < 6:
        is_bed6 = False

    if not is_bed6:
        raise ValueError("A bed6-like DataFrame is required, got {}".format(df.columns))

    columns = ["chrom", "start", "end", "name", "score", "strand"]
    bed = df[columns].copy()
    bed["chrom"] = bed["chrom"].astype(str)
    bed = bed.sort_values(["chrom", "start", "end"])

    with tempfile.NamedTemporaryFile(suffix=".bed") as f, tempfile.NamedTemporaryFile(
        "wt", suffix=".chrom.sizes"
    ) as cs:

        chromsizes.to_csv(cs, sep="\t", header=False)
        cs.flush()

        bed.to_csv(
            f.name, sep="\t", columns=columns, index=False, header=False, na_rep="nan"
        )

        p = subprocess.run(
            ["bedToBigBed", "-type={}".format(schema), f.name, cs.name, outpath],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
    return p



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import io
import os
import re

from setuptools import setup, find_packages


PKG_NAME = "bioframe"
README_PATH = "README.md"
INSTALL_DEPS_PATH = "requirements.txt"
CLASSIFIERS = [
    "Development Status :: 4 - Beta",
    "Operating System :: OS Independent",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3.9",
]


def _read(*parts, **kwargs):
    filepath = os.path.join(os.path.dirname(__file__), *parts)
    encoding = kwargs.pop("encoding", "utf-8")
    with io.open(filepath, encoding=encoding) as fh:
        text = fh.read()
    return text


def get_version(pkg_name):
    version = re.search(
        r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]',
        _read("{}/_version.py".format(pkg_name)),
        re.MULTILINE,
    ).group(1)
    return version


def get_long_description(readme_path):
    return _read(readme_path)


def get_requirements(path):
    content = _read(path)
    return [req for req in content.split("\n") if req != "" and not req.startswith("#")]


# extras_require = {
#     'docs': [
#         'Sphinx>=1.1',
#         'numpydoc>=0.5'
#     ]
# }


setup(
    name=PKG_NAME,
    author="Open2C",
    author_email="nezar@mit.edu",
    version=get_version(PKG_NAME),
    license="MIT",
    description="Pandas utilities for tab-delimited and other genomic files",
    long_description=get_long_description(README_PATH),
    long_description_content_type="text/markdown",
    url="https://github.com/open2c/bioframe",
    keywords=[
        "pandas",
        "dataframe",
        "genomics",
        "epigenomics",
        "bioinformatics",
        "intervals",
    ],
    packages=find_packages(),
    package_data={"bioframe": ["io/data/*"]},
    zip_safe=False,
    classifiers=CLASSIFIERS,
    python_requires=">=3.7",
    install_requires=get_requirements(INSTALL_DEPS_PATH),
    tests_require=["pytest"],
    setup_requires=["wheel"],
    # extras_require=extras_require,
    # entry_points={
    # }
)



import pandas as pd


def parse_gtf_attributes(attrs, kv_sep="=", item_sep=";", quotechar='"', **kwargs):
    item_lists = attrs.str.split(item_sep)
    item_lists = item_lists.apply(
        lambda items: [item.strip().split(kv_sep) for item in items]
    )
    stripchars = quotechar + " "
    item_lists = item_lists.apply(
        lambda items: [
            map(lambda x: x.strip(stripchars), item) for item in items if len(item) == 2
        ]
    )
    kv_records = item_lists.apply(dict)
    return pd.DataFrame.from_records(kv_records, **kwargs)



import itertools
import numpy as np
import pandas as pd

import matplotlib
import matplotlib.pyplot as plt

from .core import arrops

DEFAULT_FACECOLOR = "skyblue"
DEFAULT_EDGECOLOR = "dimgray"

__all__ = ["plot_intervals"]


def _plot_interval(
    start, end, level, facecolor=None, edgecolor=None, height=0.6, ax=None
):
    facecolor = DEFAULT_FACECOLOR if facecolor is None else facecolor
    edgecolor = DEFAULT_EDGECOLOR if edgecolor is None else edgecolor

    ax = plt.gca() if ax is None else ax
    ax.add_patch(
        matplotlib.patches.Rectangle(
            (start, level - height / 2),
            end - start,
            height,
            facecolor=facecolor,
            edgecolor=edgecolor,
        )
    )


def plot_intervals_arr(
    starts,
    ends,
    levels=None,
    labels=None,
    colors=None,
    xlim=None,
    show_coords=False,
    figsize=(10, 2),
):
    """
    Plot a collection of intervals.

    Parameters
    ----------
    starts, ends : np.ndarray
        A collection of intervals.

    levels : iterable or None
        The level of each interval, i.e. the y-coordinate at which the interval
        must be plotted. If None, it will be determined automatically.

    labels : str or iterable or None
        The label of each interval.

    colors : str or iterable or None.
        The color of each interval.

    xlim : (float, float) or None
        The x-span of the plot.

    show_coords : bool
        If True, plot x-ticks.

    figsize : (float, float) or None.
        The size of the figure. If None, plot within the current figure.

    """
    starts = np.asarray(starts)
    ends = np.asarray(ends)

    if figsize is not None:
        plt.figure(figsize=figsize)

    if levels is None:
        levels = arrops.stack_intervals(starts, ends)
    else:
        levels = np.asarray(levels)

    if isinstance(colors, str) or (colors is None):
        colors = itertools.cycle([colors])
    else:
        colors = itertools.cycle(colors)

    if isinstance(labels, str) or (labels is None):
        labels = itertools.cycle([labels])
    else:
        labels = itertools.cycle(labels)

    for i, (start, end, level, color, label) in enumerate(
        zip(starts, ends, levels, colors, labels)
    ):

        _plot_interval(start, end, level, facecolor=color)
        if label is not None:
            plt.text(
                (start + end) / 2,
                level,
                label,
                horizontalalignment="center",
                verticalalignment="center",
            )

    plt.ylim(-0.5, np.max(levels) + 0.5)
    if xlim is None:
        plt.xlim(-0.5, np.max(ends) + 0.5)
    else:
        plt.xlim(xlim[0], xlim[1])
    plt.gca().set_aspect(1)

    plt.gca().set_frame_on(False)
    plt.yticks([])
    if show_coords:
        pass
    else:
        plt.xticks([])


def plot_intervals(
    df,
    levels=None,
    labels=None,
    colors=None,
    xlim=None,
    show_coords=False,
    figsize=(10, 2),
):
    """
    Plot a collection of intervals, one plot per chromosome.

    Parameters
    ----------
    df : pandas.DataFrame
        A collection of intervals.

    levels : iterable or None
        The level of each interval, i.e. the y-coordinate at which the interval
        must be plotted. If None, it will be determined automatically.

    labels : str or iterable or None
        The label of each interval.

    colors : str or iterable or None.
        The color of each interval.

    xlim : (float, float) or None
        The x-span of the plot.

    show_coords : bool
        If True, plot x-ticks.

    figsize : (float, float) or None.
        The size of the figure. If None, plot within the current figure.

    """
    chrom_gb = df.groupby("chrom")
    for chrom, chrom_df in chrom_gb:
        if isinstance(levels, (list, pd.core.series.Series, np.ndarray)):
            chrom_levels = np.asarray(levels)[chrom_gb.groups[chrom].values]
        elif levels is None:
            chrom_levels = None
        else:
            raise ValueError(f"Unknown type of levels: {type(levels)}")

        if isinstance(labels, (list, pd.core.series.Series, np.ndarray)):
            chrom_labels = np.asarray(labels)[chrom_gb.groups[chrom].values]
        elif labels is None:
            chrom_labels = None
        else:
            raise ValueError(f"Unknown type of labels: {type(levels)}")

        if isinstance(colors, (list, pd.core.series.Series, np.ndarray)):
            chrom_colors = np.asarray(colors)[chrom_gb.groups[chrom].values]
        elif colors is None or isinstance(colors, str):
            chrom_colors = colors
        else:
            raise ValueError(f"Unknown type of colors: {type(colors)}")

        plot_intervals_arr(
            chrom_df.start,
            chrom_df.end,
            levels=chrom_levels,
            labels=chrom_labels,
            colors=chrom_colors,
            xlim=xlim,
            show_coords=show_coords,
            figsize=figsize,
        )
        plt.title(chrom)
